# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module contains an algorithm for detecting statistical anomalies from
   sequences of gamma spectra.
"""
import logging
from queue import Queue
from typing import Union

import numpy as np
from scipy.stats import poisson, norm


class PoissonNChannelEventDetector():
    """This class identifies sequences of anomalous spectra.

    This algorithm assumes that the background environment will be consistent over known timescales.
    I.e., this algorithm is only intended for static detector, mobile source scenarios.
    """

    def __init__(self, long_term_duration: int = 120, short_term_duration: int = 1,
                 pre_event_duration: int = 5, max_event_duration: float = 120,
                 post_event_duration: int = 1.5, tolerable_false_alarms_per_day: float = 1.0,
                 anomaly_threshold_update_interval: float = 60):
        """
        Args:
            long_term_duration: duration (in seconds) of the long-term buffer;
                this buffer contains background samples only if the correctness of the algorithm
                and environmental assumptions both hold.  A longer duration is often better with
                the trade-off eventually becoming RAM (to store all of the spectra in the buffer)
                and waiting time vs. converging on actually seeing the desired number of tolerable
                false alarms per day.
            short_term_duration: duration (in seconds) of the short-term buffer;
                this buffer can contain background and/or anomalous samples.
                This is the buffer of "most recent" samples that is compared to the long-term buffer
                in order to determine if an anomaly is present.  The basic question asked by this
                algorithm is, "Is what I see in my short-term significantly different than what I've
                seen in my long-term buffer?  If so, it's an anomaly so I'll add it to the event
                buffer.  If not, it's background so I'll add it to the long-term buffer."
                A short-term duration close to the measurement duration is recommended.
            pre_event_duration: duration (in seconds) specifying the amount of pre-event,
                background samples to include in the event result (but not summed into the event
                spectrum) in order to visualize the run-up to the anomaly.
            max_event_duration: maximum duration (in seconds) of the event buffer.
                If anomalies remain present for more than this specified amount of time, the object
                will produce an event so that something is reported.  A new event will likely then
                be started on the subsequent measurement if the anomaly is still present.
            post_event_duration: duration (in seconds) that determines the number of
                consecutive, insignificant measurements (i.e., measurements which are not considered
                anomalies) that must be observed in order to end an event.
            tolerable_false_alarms_per_day: desired maximum number of allowable false
                positive events per day for all spectrum channels.  The actual observed outcome of
                this parameter primarily depends on the long-term duration (although other
                parameters such as the short-term duration and limit update frequency also play a
                part).  Long-term duration influences the number of actual false alarms per day in
                the following ways:

                1. if the long-term buffer is not long enough to incorporate long-term
                   background fluctuations, such fluctuations can be interpreted as anomalies;
                   and
                2. if the long-term buffer is too short for the desired false alarm rate, it
                   cannot accurately model the anomaly probability distribution
                   (more specifically, to achieve a very low false alarm rate, this algorithm
                   must witness a sufficiently large number of measurements around the tail of
                   the anomaly probability distribution where the false alarm threshold exists,
                   and measurements in that region do not happen very often, hence you must
                   wait a long time to see enough).

                To give an idea of the challenge here, empirical evidence has shown that to
                accurately achieve 1 false alarm per day requires 6 or more hours of long-term
                duration for a 1024-channel, 3"x3" NaI detector at a sample interval of 0.25s.
                The unfortunate reality of this is that you are now dealing with a timescale where
                the background is likely to have a significantly different mean. So what is one to
                do? The simple answer is leave this parameter at its default, and then accept the
                reality of physics that will inevitably lead you to more false positives due to an
                unobtainable background characterization. In other words, aim high but be realistic.
            anomaly_threshold_update_interval: time (in seconds) between updates to the anomaly
                probability thresholds.  Since this calculation is expensive relative to the other
                computations going on, and because background itself usually does not significantly
                change between consecutive measurements, this parameter can be used to relax
                computation.

        """
        self._event_in_progress = False

        self._measurement_duration = None
        self._n_measurement_channels = 0
        self._post_event_duration = post_event_duration

        self._anomaly_threshold_update_interval = anomaly_threshold_update_interval
        self._anomaly_threshold_update_counter = 0
        self._pre_event_duration = pre_event_duration

        self._max_event_duration = max_event_duration
        self._tolerable_false_alarms_per_day = tolerable_false_alarms_per_day

        self._short_term_buffer = Queue()
        self._short_term_duration = short_term_duration

        self._long_term_buffer = Queue()
        self._long_term_duration = long_term_duration

        self._event_buffer = Queue()

        self._reset_stats()

    # region Properties

    @property
    def event_in_progress(self):
        """Whether an event is in progress."""
        return self._event_in_progress

    @property
    def short_term_buffer_length(self):
        """Get the number of measurements in the short term buffer (foreground)."""
        return self._short_term_max_count

    @property
    def background_percent_complete(self):
        """Get the percent fullness as a percent (not decimal)."""
        if self._long_term_max_count == 0:
            return 0
        long_term_buffer_percent_full = self._long_term_count / self._long_term_max_count * 100
        return long_term_buffer_percent_full

    @property
    def long_term_sum_norm(self):
        """Get the normalized long term buffer sum (background)."""
        return self._long_term_sum_norm

    @property
    def long_term_duration(self):
        """Get or set the duration (in seconds) of the long term buffer (background)."""
        return self._long_term_duration

    @long_term_duration.setter
    def long_term_duration(self, value):
        f_value = float(value)
        if f_value <= 0:
            raise ValueError("Long term duration must be greater than zero.")
        self._long_term_duration = f_value
        self.clear_background()

    @property
    def short_term_duration(self):
        """Get or set the duration (in seconds) of the short term buffer."""
        return self._short_term_duration

    @short_term_duration.setter
    def short_term_duration(self, value):
        f_value = float(value)
        if f_value <= 0:
            raise ValueError("Short term duration must be greater than zero.")
        self._short_term_duration = f_value

    @property
    def pre_event_duration(self):
        """Get or set the pre-event duration, e.g. the allowable time before an event
        occurs that is considered a part of the event."""
        return self._pre_event_duration

    @pre_event_duration.setter
    def pre_event_duration(self, value):
        f_value = float(value)
        if f_value <= 0:
            raise ValueError("Pre-event duration must be greater than zero.")
        self._pre_event_duration = f_value

    @property
    def max_event_duration(self):
        """Get or set the maximum event duration."""
        return self._max_event_duration

    @max_event_duration.setter
    def max_event_duration(self, value):
        f_value = float(value)
        if f_value <= 0:
            raise ValueError("Max event duration must be greater than zero.")
        self._max_event_duration = f_value

    @property
    def post_event_duration(self):
        """Get or set the post-event duration, e.g. the allowable time after an event
        occurs that is still considered part of the event."""
        return self._post_event_duration

    @post_event_duration.setter
    def post_event_duration(self, value):
        f_value = float(value)
        if f_value < 0:
            raise ValueError("Post event duration must be nonnegative.")
        self._post_event_duration = f_value

    @property
    def tolerable_false_alarms_per_day(self):
        """Get or set the number of tolerable false alarms per day."""
        return self._tolerable_false_alarms_per_day

    @tolerable_false_alarms_per_day.setter
    def tolerable_false_alarms_per_day(self, value):
        f_value = float(value)
        if f_value <= 0:
            raise ValueError("Tolerable false alarms must be greater than zero.")
        self._tolerable_false_alarms_per_day = f_value

    @property
    def limit_update_frequency(self):
        """Get or set the frequency (in Hz) which which to recompute anomaly thresholds.
        """
        return self._limit_update_frequency

    @limit_update_frequency.setter
    def limit_update_frequency(self, value):
        f_value = float(value)
        if f_value < 0:
            raise ValueError("Limit update frequency must be nonnegative.")
        self._limit_update_frequency = f_value

    # endregion

    def _reset_stats(self):
        """Reset the Event Detector stats to a zeroed, initial state."""
        self._measurement_duration = None

        self._anomalous_count = 0
        self._nonanomalous_count_to_trigger_off = 0
        self._nonanomalous_count = 0
        self._anomaly_threshold_update_counter = 0
        self._anomaly_probas = 0
        self._anomaly_proba = None
        self._anomaly_thresholds = None

        self._short_term_buffer_full = False
        self._short_term_count = 0
        self._short_term_max_count = 0
        self._short_term_sum = None

        self._long_term_buffer_full = False
        self._long_term_count = 0
        self._long_term_max_count = 0
        self._long_term_sum = None
        self._long_term_cps = 0
        self._long_term_sum_norm = 0
        self._long_term_proba_buffer = Queue()
        self._long_term_proba_stats = RunningStats(0, 0, 0)

    def clear_background(self):
        """Clear the short-term buffer, long-term buffer, and stats."""
        with self._short_term_buffer.mutex:
            self._short_term_buffer.queue.clear()
        with self._long_term_buffer.mutex:
            self._long_term_buffer.queue.clear()
        self._reset_stats()

    def add_measurement(self, measurement_id: int, measurement: Union[np.array, int],
                        duration: float, verbose=True):
        """Add the next measurement to the event detector.

        Args:
            measurement_id: unique identifier for the measurement
            measurement: 1D array-like object containing one or more channels of measured data
            duration: float representing the length of time over which the measurement was taken
            verbose: whether to print log messages

        Returns:
            if adding a spectrum concludes an event, a tuple is returned containing:
                - the channel-wise sum of all spectra that were part of the event
                - the channel-wise sum of all spectra in the background buffer,
                    normalized to match the duration of the event
                - the duration (in seconds) of the event
                - list of measurement IDs that comprise the event
            else if adding a spectrum does not conclude event, return None
        """
        if self._measurement_duration != duration:
            if verbose:
                logging.debug("Measurement duration changed")
            self.clear_background()
            self._measurement_duration = duration
            self._n_measurement_channels = len(measurement)
            self._short_term_max_count = self._short_term_duration / duration
            self._long_term_max_count = self._long_term_duration / duration
            self._nonanomalous_count_to_trigger_off = self._post_event_duration / duration
            self._sigma_deviation = \
                (self._tolerable_false_alarms_per_day / (24 * 60 * 60 / duration))
            self._long_term_proba_stats.N = self._long_term_max_count

        # Update short-term buffer
        self._short_term_buffer.put((measurement_id, measurement))
        self._short_term_buffer_full = self._short_term_count >= self._short_term_max_count
        if self._short_term_sum is not None:
            self._short_term_sum += np.array(measurement)
        else:
            self._short_term_sum = np.array(measurement)
        if self._short_term_buffer_full:
            _, oldest_short_term_measurement = self._short_term_buffer.get()
            self._short_term_sum -= oldest_short_term_measurement
        else:
            self._short_term_count += 1

        # Anomaly & event determination
        self._long_term_buffer_full = self._long_term_max_count and \
            self._long_term_count >= self._long_term_max_count
        event_result = None
        if self._long_term_buffer_full:
            # Check if it's time to update the anomaly thresholds
            time_to_update_anomaly_thresholds = not self._anomaly_thresholds or \
                self._anomaly_threshold_update_counter >= \
                self._anomaly_threshold_update_interval
            if time_to_update_anomaly_thresholds:
                # It is known that the anomaly probability distribution is more perfectly
                # modeled with a skewed normal distribution, however the number of measurements
                # required to accurately obtain the skew term is infeasible so we use a
                # gaussian for now.
                self._anomaly_thresholds = norm.ppf(
                    self._sigma_deviation,
                    self._long_term_proba_stats.average,
                    self._long_term_proba_stats.stddev
                )
                self._anomaly_threshold_update_counter = 0
            else:
                self._anomaly_threshold_update_counter += self._measurement_duration
            # Check anomaly thresholds
            self._anomaly_probas = poisson.logpmf(
                self._short_term_sum,
                self._long_term_sum_norm
            )
            self._anomaly_proba = np.sum(self._anomaly_probas)
            measurement_is_anomalous = self._anomaly_proba < self._anomaly_thresholds
            # Now determine if signficant measurements meet on_time / off_time requirements
            if measurement_is_anomalous:
                if not self._event_in_progress:
                    if verbose:
                        logging.debug(f"Event started @ {measurement_id}")
                    self._event_in_progress = True
                    self._nonanomalous_count = 0
                # event duration = anomalous count * measurement duration
                self._anomalous_count += 1
                self._event_buffer.put((measurement_id, measurement))
            elif self._event_in_progress:
                self._nonanomalous_count += 1
                anomaly_gone = self._nonanomalous_count >= self._nonanomalous_count_to_trigger_off
                event_duration = self._anomalous_count * self._measurement_duration
                event_reached_max_duration = event_duration >= self.max_event_duration
                event_over = anomaly_gone or event_reached_max_duration
                if event_over:
                    if verbose:
                        logging.debug(f"Event ended @ {measurement_id}")
                    self._event_in_progress = False
                    # Build the event tuple
                    event_measurement_ids, event_measurements = zip(
                        *[self._event_buffer.get() for _ in range(self._event_buffer.qsize())]
                    )
                    event_measurement = np.array(event_measurements).sum(axis=0)
                    self._anomalous_count = 0
                    long_term_cps = self._long_term_sum / self._long_term_duration
                    event_bg_measurement = long_term_cps * event_duration
                    gross_counts = sum(event_measurement)
                    bg_counts = sum(event_bg_measurement)
                    fg_counts = gross_counts - bg_counts
                    snr = fg_counts / bg_counts
                    is_positive_snr_event = snr > 0
                    if not is_positive_snr_event and verbose:
                        logging.debug(f"Event ending @ {measurement_id} had a SNR <= 0 ({snr:.2f})")
                    event_result = (
                        event_measurement,
                        event_bg_measurement,
                        event_duration,
                        event_measurement_ids
                    )
                else:
                    self._event_buffer.put((measurement_id, measurement))

        # Update long-term buffer
        if not self._event_in_progress:
            self._long_term_buffer.put((measurement_id, measurement))
            if self._long_term_sum is not None:
                self._long_term_sum += np.array(measurement)
            else:
                self._long_term_sum = np.array(measurement)

            if self._long_term_buffer_full:
                _, oldest_long_term_measurement = self._long_term_buffer.get()
                self._long_term_sum -= oldest_long_term_measurement
            else:
                self._long_term_count += 1

            self._long_term_cps = np.divide(self._long_term_sum,
                                            (self._long_term_count * self._measurement_duration))
            self._long_term_sum_norm = (self._long_term_cps * self._short_term_duration).clip(
                1 / self._n_measurement_channels
            )
            # Calculate some stats for thresholding
            long_term_anomaly_probas = poisson.logpmf(
                self._short_term_sum,
                self._long_term_sum_norm
            )
            long_term_anomaly_proba = np.sum(long_term_anomaly_probas)
            # Rolling Welford's algorithm
            self._long_term_proba_buffer.put(long_term_anomaly_proba)
            oldest_long_term_proba = 0
            if self._long_term_buffer_full:
                oldest_long_term_proba = self._long_term_proba_buffer.get()
            self._long_term_proba_stats.update(long_term_anomaly_proba, oldest_long_term_proba)

        return event_result


class RunningStats(object):
    """Helper for calculating stats in a rolling fashion."""

    def __init__(self, window_size, average, variance):
        """
        Args:
            window_size: size of the window across which the probabilities are calculated
            average: average within the window
            variance: variance within the window
        """
        self.N = window_size
        self.average = average
        self.variance = variance
        self.stddev = np.sqrt(variance)

    def update(self, new, old):
        """Update the average, variance, and stddev for the RunningStats.

        Args:
            new: new value to include in window
            old: oldest value in window (this must be externally tracked - queue)
        """
        oldavg = self.average
        newavg = oldavg + (new - old) / self.N
        self.average = newavg
        self.variance += (new - old) * (new - newavg + old - oldavg) / (self.N - 1)
        self.stddev = np.sqrt(self.variance)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module contains functionality shared across all PyRIID models."""
import json
import os
import uuid
from abc import abstractmethod
from enum import Enum

import numpy as np
import tensorflow as tf
import tf2onnx
from keras.models import Model
from keras.utils import get_custom_objects

import riid
from riid.data.labeling import label_to_index_element
from riid.data.sampleset import SampleSet, SpectraState
from riid.losses import mish
from riid.metrics import multi_f1, single_f1

get_custom_objects().update({
    "multi_f1": multi_f1,
    "single_f1": single_f1,
    "mish": mish,
})


class ModelInput(int, Enum):
    """Enumerates the potential input sources for a model."""
    GrossSpectrum = 0
    BackgroundSpectrum = 1
    ForegroundSpectrum = 2


class PyRIIDModel:
    """Base class for PyRIID models."""

    def __init__(self, *args, **kwargs):
        self._info = {}
        self._temp_file_path = "temp_model.json"
        self._custom_objects = {}
        self._initialize_info()

    @property
    def seeds(self):
        return self._info["seeds"]

    @seeds.setter
    def seeds(self, value):
        self._info["seeds"] = value

    @property
    def info(self):
        return self._info

    @info.setter
    def info(self, value):
        self._info = value

    @property
    def target_level(self):
        return self._info["target_level"]

    @target_level.setter
    def target_level(self, value):
        if value in SampleSet.SOURCES_MULTI_INDEX_NAMES:
            self._info["target_level"] = value
        else:
            msg = (
                f"Target level '{value}' is invalid.  "
                f"Acceptable levels: {SampleSet.SOURCES_MULTI_INDEX_NAMES}"
            )
            raise ValueError(msg)

    @property
    def model(self) -> Model:
        return self._model

    @model.setter
    def model(self, value: Model):
        self._model = value

    @property
    def model_id(self):
        return self._info["model_id"]

    @model_id.setter
    def model_id(self, value):
        self._info["model_id"] = value

    @property
    def model_inputs(self):
        return self._info["model_inputs"]

    @model_inputs.setter
    def model_inputs(self, value):
        self._info["model_inputs"] = value

    @property
    def model_outputs(self):
        return self._info["model_outputs"]

    @model_outputs.setter
    def model_outputs(self, value):
        self._info["model_outputs"] = value

    def get_model_outputs_as_label_tuples(self):
        return [
            label_to_index_element(v, self.target_level) for v in self.model_outputs
        ]

    def _get_model_dict(self) -> dict:
        model_json = self.model.to_json()
        model_dict = json.loads(model_json)
        model_weights = self.model.get_weights()
        model_dict = {
            "info": self._info,
            "model": model_dict,
            "weights": model_weights,
        }
        return model_dict

    def _get_model_str(self) -> str:
        model_dict = self._get_model_dict()
        model_str = json.dumps(model_dict, indent=4, cls=PyRIIDModelJsonEncoder)
        return model_str

    def _initialize_info(self):
        init_info = {
            "model_id": str(uuid.uuid4()),
            "model_type": self.__class__.__name__,
            "normalization": SpectraState.Unknown,
            "pyriid_version": riid.__version__,
        }
        self._update_info(**init_info)

    def _update_info(self, **kwargs):
        self._info.update(kwargs)

    def _update_custom_objects(self, key, value):
        self._custom_objects.update({key: value})

    def load(self, model_path: str):
        """Load the model from a path.

        Args:
            model_path: path from which to load the model.
        """
        if not os.path.exists(model_path):
            raise ValueError("Model file does not exist.")

        with open(model_path) as fin:
            model = json.load(fin)

        model_str = json.dumps(model["model"])
        self.model = tf.keras.models.model_from_json(model_str, custom_objects=self._custom_objects)
        self.model.set_weights([np.array(x) for x in model["weights"]])
        self.info = model["info"]

    def save(self, model_path: str, overwrite=False):
        """Save the model to a path.

        Args:
            model_path: path at which to save the model.
            overwrite: whether to overwrite an existing file if it already exists.

        Raises:
            `ValueError` when the given path already exists
        """
        if os.path.exists(model_path) and not overwrite:
            raise ValueError("Model file already exists.")

        model_str = self._get_model_str()
        with open(model_path, "w") as fout:
            fout.write(model_str)

    def to_onnx(self, model_path: str = None, **tf2onnx_kwargs: dict):
        """Convert the model to an ONNX model.

        Args:
            model_path: path at which to save the model
            tf2onnx_kwargs: additional kwargs to pass to the conversion
        """
        if not model_path.endswith(riid.ONNX_MODEL_FILE_EXTENSION):
            raise ValueError(f"ONNX file path must end with {riid.ONNX_MODEL_FILE_EXTENSION}")
        if os.path.exists(model_path):
            raise ValueError("Model file already exists.")

        tf2onnx.convert.from_keras(
            self.model,
            output_path=model_path,
            **tf2onnx_kwargs
        )

    def to_tflite(self, model_path: str, quantize: bool = False, prune: bool = False):
        """Convert the model to a TFLite model and optionally applying quantization or pruning.

        Args:
            model_path: file path at which to save the model
            quantize: whether to apply quantization
            prune: whether to apply pruning
        """
        if not model_path.endswith(riid.TFLITE_MODEL_FILE_EXTENSION):
            raise ValueError(f"TFLite file path must end with {riid.TFLITE_MODEL_FILE_EXTENSION}")
        if os.path.exists(model_path):
            raise ValueError("Model file already exists.")

        optimizations = []
        if quantize:
            optimizations.append(tf.lite.Optimize.DEFAULT)
        if prune:
            optimizations.append(tf.lite.Optimize.EXPERIMENTAL_SPARSITY)

        converter = tf.lite.TFLiteConverter.from_keras_model(self.model)
        converter.optimizations = optimizations
        tflite_model = converter.convert()

        with open(model_path, "wb") as fout:
            fout.write(tflite_model)

    @abstractmethod
    def fit(self):
        pass

    @abstractmethod
    def predict(self):
        pass


class PyRIIDModelJsonEncoder(json.JSONEncoder):
    """Custom JSON encoder for saving models.
    """
    def default(self, o):
        """Converts certain types to JSON-compatible types.
        """
        if isinstance(o, np.ndarray):
            return o.tolist()
        elif isinstance(o, np.float32):
            return o.astype(float)

        return super().default(o)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module provides tools for handling data related to the multi-lab Algorithm Improvement
Program Team (AIPT).
"""
import os
from pathlib import Path
from typing import List

import pandas as pd

from riid import SAMPLESET_HDF_FILE_EXTENSION
from riid.data.converters import _validate_and_create_output_dir
from riid.data.sampleset import SampleSet

ELEMENT_IDS_PER_FILE = [0, 1, 2, 3]
DEFAULT_ECAL = [
    -3.00000,
    3010.00000,
    150.00000,
    0.00000,
    0.00000,
]


def _element_to_ss(data_df: pd.DataFrame, eid: int, description) -> SampleSet:
    if eid not in ELEMENT_IDS_PER_FILE:
        msg = (
            f"Element #{eid} is invalid. "
            "The available options are: {ELEMENT_IDS_PER_FILE}"
        )
        raise ValueError(msg)

    ss = SampleSet()
    ss.spectra = data_df[f"spectrum-channels{eid}"]\
        .str\
        .split(",", expand=True)\
        .astype(int)
    ss.info.live_time = data_df[f"spectrum-lt{eid}"] / 1000
    ss.info.real_time = data_df[f"spectrum-rt{eid}"] / 1000
    ss.info.total_counts = data_df[f"gc{eid}"]
    ss.info.neutron_counts = data_df["nc0"]
    ss.info.ecal_order_0 = DEFAULT_ECAL[0]
    ss.info.ecal_order_1 = DEFAULT_ECAL[1]
    ss.info.ecal_order_2 = DEFAULT_ECAL[2]
    ss.info.ecal_order_3 = DEFAULT_ECAL[3]
    ss.info.ecal_low_e = DEFAULT_ECAL[4]
    ss.info.timestamp = data_df["utc-time"]
    ss.info.description = description
    ss.info["latitude"] = data_df["latitude"]
    ss.info["longitude"] = data_df["longitude"]
    ss.info["is_in_zone"] = data_df["is-in-zone"]
    ss.info["is_closest_approach"] = data_df["is-closest-approach"]
    ss.info["is_source_present"] = data_df["is-source-present"]

    detector_name = f"{data_df['detector'].unique()[0]}.{eid}"
    ss.detector_info = {
        "name": detector_name
    }

    return ss


def aipt_file_to_ss_list(file_path: str) -> List[SampleSet]:
    """Process an AIPT CSV file into a list of SampleSets.

    Each file contains a series of spectra for multiple detectors running simultaneously.
    As such, each `SampleSet` in the list returned by this function represents the data
    collected by each detector.
    Each row of each `SampleSet` represents a measurement from each detector at the same
    moment in time.
    As such, after calling this function, you might consider summing all four spectra at
    each timestep into a single spectrum as another processing step.

    Args:
        file_path: file path of the CSV file

    Returns:
        List of `SampleSet`s each containing a series of spectra for a single run
    """
    data_df = pd.read_csv(file_path, header=0, sep="\t")
    base_description = os.path.splitext(os.path.basename(file_path))[0]

    ss_list = []
    for eid in ELEMENT_IDS_PER_FILE:
        description = f"{base_description}_{eid}"
        ss = _element_to_ss(data_df, eid, description)
        ss_list.append(ss)

    return ss_list


def convert_and_save(input_file_path: str, output_dir: str = None,
                     skip_existing: bool = True, **kwargs):
    """Convert AIPT file to SampleSet and save as HDF.

    Output file will have same name but appended with a detector identifier
    and having a different extension.

    Args:
        input_file_path: file path of the CSV file
        output_dir: alternative directory in which to save HDF files
            (defaults to `input_file_path` parent if not provided)
        skip_existing: whether to skip conversion if the file already exists
        kwargs: keyword args passed to `aipt_file_to_ss_list()` (not currently used)
    """
    input_path = Path(input_file_path)
    if not output_dir:
        output_dir = input_path.parent
    _validate_and_create_output_dir(output_dir)
    output_file_paths = [
        os.path.join(output_dir, input_path.stem + f"-{i}{SAMPLESET_HDF_FILE_EXTENSION}")
        for i in ELEMENT_IDS_PER_FILE
    ]
    all_output_files_exist = all([os.path.exists(p) for p in output_file_paths])
    if skip_existing and all_output_files_exist:
        return

    ss_list = aipt_file_to_ss_list(input_file_path, **kwargs)
    for output_file_path, ss in zip(output_file_paths, ss_list):
        ss.to_hdf(output_file_path)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module contains utility functions for managing ground truth information."""
import logging
import re

BACKGROUND_LABEL = "Background"
NO_SEED = "Unknown"
NO_ISOTOPE = "Unknown"
NO_CATEGORY = "Uncategorized"
CATEGORY_ISOTOPES = {
    "Fission Product": {
        "severity": 3,
        "isotopes": [
            "Ag112",
            "As78",
            "Ba139",
            "Ce143",
            "I132",
            "I133",
            "I134",
            "I135",
            "Kr85m",
            "Kr87",
            "La140",
            "La142",
            "Nd149",
            "Pm150",
            "Rh105",
            "Ru105",
            "Sb115",
            "Sb129",
            "Sr91",
            "Sr92",
            "Te132",
            "Y93",
            "Y91m",
            "Zr95",
        ]
    },
    "Industrial": {
        "severity": 2,
        "isotopes": [
            "Am241",
            "Ba133",
            "Ba140",
            "Bi207",
            "Cf249",
            "Cf250",
            "Cf251",
            "Cf252",
            "Cm244",
            "Co57",
            "Co60",
            "Cs137",
            "Eu152",
            "Eu154",
            "H3",
            "Ho166m",
            "Ir192",
            "Na22",
            "P32",
            "P33",
            "Po210",
            "Se75",
            "Sr90",
            "Tc99",
            "Y88",
        ],
    },
    "Medical": {
        "severity": 3,
        "isotopes": [
            "F18",
            "Ga67",
            "Ga68",
            "Ge68",
            "I123",
            "I124",
            "I125",
            "I129",
            "I131",
            "In111",
            "Lu177m",
            "Mo99",
            "Pd103",
            "Ra223",
            "Rb82",
            "Sm153",
            "Tc99m",
            "Tl201",
            "Xe133",
        ],
    },
    "NORM": {
        "severity": 1,
        "isotopes": [
            "Cosmic",
            "K40",
            "Pb210",
            "Ra226",
            "Th232",
        ],
    },
    "SNM": {
        "severity": 4,
        "isotopes": [
            "Np237",
            "Pu238",
            "Pu239",
            "U232",
            "U233",
            "U235",
            "U237",
            # 16,000 years from now, once the chemically separated uranium has equilibrated
            # with Ra226, then we will need to reconsider U238's categorization.
            "U238",
        ],
    },
}
ISOTOPES = sum(
    [c["isotopes"] for c in CATEGORY_ISOTOPES.values()],
    []
)  # Concatenating the lists of isotopes into one list
SEED_TO_ISOTOPE_SPECIAL_CASES = {
    "ThPlate": "Th232",
    "ThPlate+Thxray,10uC": "Th232",
    "fiestaware": "U238",
    "Uxray,100uC": "U238",
    "DUOxide": "U238",
    "ShieldedDU": "U238",
    "modified_berpball": "Pu239",
    "10 yr WGPu in Fe": "Pu239",
    "1gPuWG_0.5yr,3{an=10,ad=5}": "Pu239",
    "pu239_1yr": "Pu239",
    "pu239_5yr": "Pu239",
    "pu239_10yr": "Pu239",
    "pu239_25yr": "Pu239",
    "pu239_50yr": "Pu239",
    "1kg HEU + 800uCi Cs137": "U235",
    "WGPu + Cs137": "Pu239",
    "HEU": "U235",
    "HEU + Tc99m": "U235",
    "DU": "U238",
    "WGPu": "Pu239",
    "PuWG": "Pu239",
    "RTG": "Pu238",
    "PotassiumInSoil": "K40",
    "UraniumInSoil": "Ra226",
    "ThoriumInSoil": "Th232",
    "Cosmic": "Cosmic",
}


def _find_isotope(seed: str, verbose=True):
    """Attempt to find the category for the given seed.

    Args:
        seed: string containing the isotope name
        verbose: whether log warnings

    Returns:
        Isotope if found, otherwise NO_ISOTOPE
    """
    if seed.lower() == BACKGROUND_LABEL.lower():
        return BACKGROUND_LABEL
    if seed == NO_ISOTOPE:
        return NO_ISOTOPE

    isotopes = []
    for i in ISOTOPES:
        if i in seed:
            isotopes.append(i)

    n_isotopes = len(isotopes)
    if n_isotopes > 1:
        # Use the longest matching isotope (handles sources strings for things like Tc99 vs Tc99m)
        chosen_match = max(isotopes)
        if verbose:
            logging.warning((
                f"Found multiple isotopes whose names are subsets of '{seed}';"
                f" '{chosen_match}' was chosen."
            ))
        return chosen_match
    elif n_isotopes == 0:
        return NO_ISOTOPE
    else:
        return isotopes[0]


def _find_category(isotope: str):
    """Attempt to find the category for the given isotope.

    Args:
        isotope: string containing the isotope name

    Returns:
        Category if found, otherwise NO_CATEGORY
    """
    if isotope.lower() == BACKGROUND_LABEL.lower():
        return BACKGROUND_LABEL
    if isotope == NO_CATEGORY:
        return NO_CATEGORY

    categories = []
    for c, v in CATEGORY_ISOTOPES.items():
        c_severity = v["severity"]
        c_isotopes = v["isotopes"]
        for i in c_isotopes:
            if i in isotope:
                categories.append((c, c_severity))

    n_categories = len(categories)
    if n_categories > 1:
        return max(categories, key=lambda x: x[1])[0]
    elif n_categories == 0:
        return NO_CATEGORY
    else:
        return categories[0][0]


def label_to_index_element(label_val: str, label_level="Isotope", verbose=False) -> tuple:
    """Try to map a label to a tuple for use in `DataFrame` `MultiIndex` columns.

    Depending on the level of the label value, you will get different tuple:

    | Label Level  | Resulting Tuple          |
    |:-------------|:-------------------------|
    |Seed          |(Category, Isotope, Seed) |
    |Isotope       |(Category, Isotope)       |
    |Category      |(Category,)               |

    Args:
        label_val: part of the label (Category, Isotope, Seed)
            from which to map the other two label values, if possible
        label_level: level of the part of the label provided, e.g,
            "Category", "Isotope", or "Seed"

    Returns:
        Tuple containing the Category, Isotope, and/or Seed values identified
        for the old label format.
    """

    old_label = label_val.strip()
    # Some files use 'background', others use 'Background'.
    if old_label.lower() == BACKGROUND_LABEL.lower():
        old_label = BACKGROUND_LABEL

    if label_level == "Category":
        return (old_label,)

    if label_level == "Isotope":
        category = _find_category(old_label)
        return (category, old_label)

    if label_level == "Seed":
        special_cases = SEED_TO_ISOTOPE_SPECIAL_CASES.keys()
        exact_match = old_label in special_cases
        first_partial_match = None
        if not exact_match:
            first_partial_match = next((x for x in special_cases if x in old_label), None)
        if exact_match:
            isotope = SEED_TO_ISOTOPE_SPECIAL_CASES[old_label]
        elif first_partial_match:
            isotope = SEED_TO_ISOTOPE_SPECIAL_CASES[first_partial_match]
        else:
            isotope = _find_isotope(old_label, verbose)
        category = _find_category(isotope)
        return (category, isotope, old_label)


def isotope_name_is_valid(isotope: str):
    """Validate whether the given string contains a properly formatted radioisotope name.

    Note that this function does NOT look up a string to determine if the string corresponds
    to a radioisotope that actually exists, it just checks the format.

    The regular expression used by this function looks for the following (in order):

    - 1 capital letter
    - 0 to 1 lowercase letters
    - 1 to 3 numbers
    - an optional "m" for metastable

    Examples of properly formatted isotope names:

    - Y88
    - Ba133
    - Ho166m

    Args:
        isotope: string containing the isotope name

    Returns:
        Bool representing whether the name string is valid
    """
    validator = re.compile(r"^[A-Z]{1}[a-z]{0,1}[0-9]{1,3}m?$")
    other_valid_names = ["fiestaware"]
    match = validator.match(isotope)
    is_valid = match is not None or \
        isotope.lower() in other_valid_names
    return is_valid



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""
.. include:: ../README.md
"""
import logging
import os
import sys

from importlib.metadata import version

HANDLER = logging.StreamHandler(sys.stdout)
logging.root.addHandler(HANDLER)
logging.root.setLevel(logging.DEBUG)
MPL_LOGGER = logging.getLogger("matplotlib")
MPL_LOGGER.setLevel(logging.WARNING)
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"

SAMPLESET_HDF_FILE_EXTENSION = ".h5"
SAMPLESET_JSON_FILE_EXTENSION = ".json"
PCF_FILE_EXTENSION = ".pcf"
ONNX_MODEL_FILE_EXTENSION = ".onnx"
TFLITE_MODEL_FILE_EXTENSION = ".tflite"
RIID = "riid"

__version__ = version(RIID)

__pdoc__ = {
    "riid.data.synthetic.seed.SeedMixer.__call__": True,
    "riid.data.synthetic.passby.PassbySynthesizer._generate_single_passby": True,
    "riid.data.sampleset.SampleSet._channels_to_energies": True,
}



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module provides custom model metrics."""
import numpy as np
import sklearn

from riid.data.sampleset import SampleSet


def multi_f1(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Calculate a measure of the F1 score of two tensors.

    Values for `y_true` and `y_pred` are assumed to sum to 1.

    Args:
        y_true: list of ground truth
        y_pred: list of predictions to compare against the ground truth

    Returns:
        Multi F1-score value(s)
    """
    from keras import backend as K

    diff = y_true - y_pred
    negs = K.clip(diff, -1.0, 0.0)
    false_positive = -K.sum(negs, axis=-1)
    true_positive = 1.0 - false_positive

    return K.mean(true_positive)


def single_f1(y_true: np.ndarray, y_pred: np.ndarray):
    """Compute the weighted F1 score for the maximum prediction and maximum ground truth.

    Values for `y_true` and `y_pred` are assumed to sum to 1.

    Args:
        y_true: list of ground truth
        y_pred: list of predictions to compare against the ground truth

    Returns:
        F1-score value(s)
    """
    import tensorflow as tf
    from keras import backend as K

    a = tf.dtypes.cast(y_true == K.max(y_true, axis=1)[:, None], tf.float32)
    b = tf.dtypes.cast(y_pred == K.max(y_pred, axis=1)[:, None], tf.float32)

    TP_mat = tf.dtypes.cast(K.all(tf.stack([a, b]), axis=0), tf.float32)
    FP_mat = tf.dtypes.cast(K.all(tf.stack([a != b, b == 1]), axis=0), tf.float32)
    FN_mat = tf.dtypes.cast(K.all(tf.stack([a != b, a == 1]), axis=0), tf.float32)

    TPs = K.sum(TP_mat, axis=0)
    FPs = K.sum(FP_mat, axis=0)
    FNs = K.sum(FN_mat, axis=0)

    F1s = 2 * TPs / (2*TPs + FNs + FPs + tf.fill(tf.shape(TPs), K.epsilon()))

    support = K.sum(a, axis=0)
    f1 = K.sum(F1s * support) / K.sum(support)
    return f1


def harmonic_mean(x, y):
    """Compute the harmonic mean of two same-dimensional arrays.

    Used to compute F1 score:

    ```
    f1_score = harmonic_mean(precision, recall)
    ```

    Args:
        x (array-like): numeric or array_like of numerics
        y (array-like): numeric or array_like of numerics matching the shape/type of `x`

    Returns:
        Array-like harmonic mean of `x` and `y`
    """
    return 2 * x * y / (x + y)


def precision_recall_curve(ss: SampleSet, smooth: bool = True, multiclass: bool = None,
                           include_micro: bool = True, target_level: str = "Isotope",
                           minimum_contribution: float = 0.01):
    """Similar to `sklearn.metrics.precision_recall_curve`, however, this function
    computes the precision and recall for each class, and supports both multi-class
    and multi-label problems.

    The reason this is necessary is that in multi-class problems, for a single sample,
    all predictions are discarded except for the argmax.

    Args:
        ss: `SampleSet` that predictions were generated on
        smooth: if True, precision is smoothing is applied to make a monotonically
            decreasing precision function
        multiclass: set to True if this is a multi-class (i.e. y_true is one-hot) as
            opposed to multi-label (i.e. labels are not mutually exclusive). Ff True,
            predictions will be masked such that non-argmax predictions are set to zero
            (this prevents inflating the precision by continuing past a point that could
            be pragmatically useful). Furthermore, in the multiclass case the recall is
            not guaranteed to reach 1.0.
        include_micro: if True, compute an additional precision and recall for the
            micro-average across all labels and put it under entry `"micro"`
        target_level: `SampleSet.sources` and `SampleSet.prediction_probas` column level to use
        minimum_contribution: threshold for a source to be considered a ground truth positive
            label. if this is set to `None` the raw mixture ratios will be used as y_true.

    Returns:
        precision (dict): dict with keys for each label and values that are the
            monotonically increasing precision values at each threshold
        recall (dict): dict with keys for each label and values that are the
            monotonically decreasing recall values at each threshold
        thresholds (dict): dict with keys for each label and values that are the
            monotonically increasing thresholds on the decision function used to compute
            precision and recall

    References:
        - [Precision smoothing](
           https://jonathan-hui.medium.com/map-mean-average-precision-for-object-detection-45c121a31173)

    """
    y_true = ss.sources.groupby(axis=1, level=target_level, sort=False).sum()
    if minimum_contribution is not None:
        y_true = (y_true > minimum_contribution).astype(int)
    y_pred = ss.prediction_probas.groupby(axis=1, level=target_level, sort=False).sum()

    # switch from pandas to numpy
    labels = y_true.columns
    n_classes = len(labels)
    y_true = y_true.values.copy()
    y_pred = y_pred.values.copy()

    if y_pred.shape != y_true.shape:
        raise ValueError(
            f"Shape mismatch between truth and predictions, "
            f"{y_true.shape} != {y_pred.shape}. "
            f"It is possible that the `target_level` is incorrect."
        )

    if multiclass is None:
        # infer whether multi-class or multi-label
        multiclass = not np.any(y_true.sum(axis=1) != 1)

    # drop nans
    notnan = ~np.isnan(y_pred).any(axis=1)
    y_true = y_true[notnan, :]
    y_pred = y_pred[notnan, :]

    y_pred_min = None
    if multiclass:
        # shift predictions to force positive
        if np.any(y_pred < 0):
            y_pred_min = y_pred.min()
            y_pred -= y_pred_min

        # mask the predictions by argmax
        pred_mask = np.eye(n_classes)[np.argmax(y_pred, axis=1)]
        # mask the predictions
        y_pred *= pred_mask

    precision = dict()
    recall = dict()
    thresholds = dict()
    for i, label in enumerate(labels):
        precision[label], recall[label], thresholds[label] = _pr_curve(
            y_true[:, i], y_pred[:, i], multiclass=multiclass, smooth=smooth
        )

    if include_micro:
        # A "micro-average": quantifying score on all classes jointly
        precision["micro"], recall["micro"], thresholds["micro"] = _pr_curve(
            y_true.ravel(), y_pred.ravel(), multiclass=multiclass, smooth=smooth
        )

    # un-shift thresholds if predictions were shifted
    if y_pred_min is not None:
        thresholds = {k: v + y_pred_min for k, v in thresholds.items()}

    return precision, recall, thresholds


def average_precision_score(precision, recall):
    """Compute the average precision (area under the curve) for each precision/recall
    pair.

    Args:
        precision (dict): return value of `ctutil.evaluation.precision_recall_curve()`
        recall (dict): return value of `ctutil.evaluation.precision_recall_curve()`

    Returns:
        (dict): average precision values (float) for each label in precision/recall
    """
    return {label: _integrate(recall[label], precision[label]) for label in recall}


def _step(x):
    """Compute the right going maximum of `x` and all previous values of `x`.

    Args:
        x (array-like): 1D array to process

    Returns:
        (array-like): right-going maximum of `x`

    """
    y = np.array(x)
    for i in range(1, len(y)):
        y[i] = max(y[i], y[i - 1])
    return y


def _integrate(x, y, y_left=True):
    """Integrate an (x, y) function pair.

    Args:
        x (array-like): 1D array of x values
        y (array-like): 1D array of y values
        y_left: if true, omit the last value of y, else, omit the first value

    Returns:
        (float): integrated "area under the curve"

    """
    delta_x = x[1:] - x[:-1]
    y_trimmed = y[:-1] if y_left else y[1:]
    return np.abs(np.sum(delta_x * y_trimmed))


def _pr_curve(y_true, y_pred, multiclass, smooth):
    precision, recall, thresholds = sklearn.metrics.precision_recall_curve(
        y_true, y_pred
    )

    if smooth:
        precision = _step(precision)

    if multiclass:
        # remove the point where threshold=0 and recall=1
        precision = precision[1:]
        recall = recall[1:]
        thresholds = thresholds[1:]

    return precision, recall, thresholds


def build_keras_semisupervised_metric_func(keras_metric_func, activation_func,
                                           n_labels):
    def metric_func(y_true, y_pred):
        return keras_metric_func(y_true[:, :n_labels], activation_func(y_pred))
    metric_func.__name__ = keras_metric_func.__name__

    return metric_func



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module contains implementations of the ARAD deep learning architecture."""
from typing import List

import pandas as pd
import tensorflow as tf
from keras.activations import sigmoid, softplus
from keras.callbacks import EarlyStopping, ReduceLROnPlateau
from keras.initializers import GlorotNormal, HeNormal
from keras.layers import (BatchNormalization, Concatenate, Conv1D,
                          Conv1DTranspose, Dense, Dropout, Flatten, Input,
                          MaxPool1D, Reshape, UpSampling1D)
from keras.models import Model
from keras.regularizers import L1L2, L2
from scipy.spatial.distance import jensenshannon
from scipy.stats import entropy

from riid.data.sampleset import SampleSet, SpectraState
from riid.losses import jensen_shannon_distance, mish
from riid.models import PyRIIDModel


class ARADv1TF(Model):
    """TensorFlow Implementation of ARAD v1.

    This implementation was built based on our interpretation of the methods outlined in the
    following paper and may not fully represent the original implementation:

    - Ghawaly Jr, James M. "A Datacentric Algorithm for Gamma-ray Radiation Anomaly Detection
      in Unknown Background Environments." (2020).
    """
    def __init__(self, latent_dim: int = 5):
        """
        Args:
            latent_dim: dimension of internal latent represention.
                5 was the final one in the paper, but 4 to 8 were found to work well.
        """
        super().__init__()

        input_size = (128, 1)
        # Encoder
        b1_config = (
            (5, 1, 32),
            (6, 2, 16),
            (6, 2, 8),
            (4, 2, 4),
            (1, 1, 2),
        )
        b2_config = (
            (10, 2, 32),
            (6, 2, 16),
            (4, 2, 8),
            (1, 1, 4),
            (1, 1, 2),
        )
        b3_config = (
            (3, 1, 32),
            (6, 2, 16),
            (3, 2, 8),
            (4, 2, 4),
            (1, 1, 2),
        )
        encoder_input = Input(shape=input_size, name="encoder_input")
        b1 = self._get_branch(encoder_input, b1_config, 0.1, "softplus", "B1", 5)
        b2 = self._get_branch(encoder_input, b2_config, 0.1, "softplus", "B2", 5)
        b3 = self._get_branch(encoder_input, b3_config, 0.1, "softplus", "B3", 5)
        x = Concatenate(axis=1)([b1, b2, b3])
        x = Reshape((15,), name="reshape")(x)
        latent_space = Dense(
            units=latent_dim,
            name="D1_latent_space",
            activation=softplus
        )(x)
        encoder_output = BatchNormalization(name="D1_batch_norm")(latent_space)
        encoder = Model(encoder_input, encoder_output, name="encoder")

        # Decoder
        decoder_input = Input(shape=(latent_dim,), name="decoder_input")
        x = Dense(units=40, name="D2", activation=softplus)(decoder_input)
        x = Dropout(rate=0.1, name="D2_dropout")(x)
        decoder_output = Dense(
            units=128,
            activation=softplus,
            name="D3"
        )(x)
        decoder = Model(decoder_input, decoder_output, name="decoder")

        # Autoencoder
        encoded_spectrum = encoder(encoder_input)
        decoded_spectrum = decoder(encoded_spectrum)
        autoencoder = Model(encoder_input, decoded_spectrum, name="autoencoder")

        def logcosh_with_kld_penalty(input_spectrum, decoded_spectrum,
                                     latent_space, sparsities,
                                     penalty_weight=0.5):
            squeezed_input = tf.squeeze(input_spectrum)
            logcosh_loss = tf.keras.losses.log_cosh(squeezed_input, decoded_spectrum)
            kld_loss = tf.keras.losses.kld(sparsities, latent_space)
            loss = logcosh_loss + penalty_weight * kld_loss
            return loss

        sparsities = [0.001] * latent_dim
        loss_func = logcosh_with_kld_penalty(
            encoder_input,
            decoded_spectrum,
            latent_space,
            sparsities,
        )
        autoencoder.add_loss(loss_func)

        self.encoder = encoder
        self.decoder = decoder
        self.autoencoder = autoencoder

    def _get_branch(self, input_layer, config, dropout_rate, activation, branch_name, dense_units):
        x = input_layer
        for i, (kernel_size, strides, filters) in enumerate(config, start=1):
            layer_name = f"{branch_name}_C{i}"
            x = Conv1D(kernel_size=kernel_size, strides=strides, filters=filters,
                       activation=activation, name=layer_name)(x)
            x = BatchNormalization(name=f"{layer_name}_batch_norm")(x)
            x = Dropout(rate=dropout_rate, name=f"{layer_name}_dropout")(x)
        x = Flatten(name=f"{branch_name}_flatten")(x)
        x = Dense(
            units=dense_units,
            name=f"{branch_name}_D1",
            activation=activation
        )(x)
        x = BatchNormalization(name=f"{branch_name}_batch_norm")(x)
        return x

    def call(self, x):
        decoded = self.autoencoder(x)
        return decoded


class ARADv2TF(Model):
    """TensorFlow Implementation of ARAD v2.

    This implementation of ARAD was built based on our interpretation of the methods outlined in the
    following paper and may not fully represent the original implementation:

    - Ghawaly Jr, James M., et al. "Characterization of the Autoencoder Radiation Anomaly Detection
      (ARAD) model." Engineering Applications of Artificial Intelligence 111 (2022): 104761.
    """
    def __init__(self, latent_dim: int = 8):
        """
        Args:
            latent_dim: dimension of internal latent represention.
                5 was the final one in the paper, but 4 to 8 were found to work well.
        """
        super().__init__()

        input_size = (128, 1)
        # Encoder
        config = (
            (7, 1, 8, 2),
            (5, 1, 8, 2),
            (3, 1, 8, 2),
            (3, 1, 8, 2),
            (3, 1, 8, 2),
        )
        encoder_input = Input(shape=input_size, name="encoder_input")
        x = encoder_input
        for i, (kernel_size, strides, filters, max_pool_size) in enumerate(config, start=1):
            conv_name = f"conv{i}"
            x = Conv1D(
                kernel_size=kernel_size,
                strides=strides,
                filters=filters,
                padding="same",
                activation=mish,
                kernel_regularizer=L1L2(l1=1e-3, l2=1e-3),
                bias_regularizer=L2(l2=1e-3),
                kernel_initializer=HeNormal,
                name=conv_name)(x)
            x = BatchNormalization(name=f"{conv_name}_batch_norm")(x)
            pool_name = f"MP{i}"
            x = MaxPool1D(pool_size=max_pool_size, name=pool_name)(x)
            x = BatchNormalization(name=f"{pool_name}_batch_norm")(x)

        x = Flatten(name="flatten")(x)
        x = Dense(
            units=latent_dim,
            activation=mish,
            kernel_regularizer=L1L2(l1=1e-3, l2=1e-3),
            bias_regularizer=L2(l2=1e-3),
            kernel_initializer=HeNormal,
            name="D1"
        )(x)
        encoder_output = BatchNormalization(name="D1_batch_norm")(x)
        encoder = Model(encoder_input, encoder_output, name="encoder")

        # Decoder
        decoder_input = Input(shape=latent_dim, name="decoder_input")
        x = Dense(units=32, name="D2", activation=mish)(decoder_input)
        x = BatchNormalization(name="D2_batch_norm")(x)
        x = Reshape((4, 8), name="reshape")(x)
        reversed_config = enumerate(reversed(config[1:]), start=1)
        for i, (kernel_size, strides, filters, max_pool_size) in reversed_config:
            upsample_name = f"US{i}"
            x = UpSampling1D(size=max_pool_size, name=upsample_name)(x)
            x = BatchNormalization(name=f"{upsample_name}_batch_norm")(x)
            conv_name = f"tconv{i}"
            x = Conv1D(
                kernel_size=kernel_size,
                strides=strides,
                filters=filters,
                padding="same",
                activation=mish,
                kernel_regularizer=L1L2(l1=1e-3, l2=1e-3),
                bias_regularizer=L2(l2=1e-3),
                kernel_initializer=HeNormal,
                name=conv_name)(x)
            x = BatchNormalization(name=f"{conv_name}_batch_norm")(x)

        i += 1
        upsample_name = f"US{i}"
        x = UpSampling1D(size=max_pool_size, name=upsample_name)(x)
        x = BatchNormalization(name=f"{upsample_name}_batch_norm")(x)
        x = Conv1DTranspose(
            kernel_size=7,
            strides=1,
            filters=1,
            padding="same",
            activation=sigmoid,
            kernel_initializer=GlorotNormal,
            name=f"tconv{i}"
        )(x)
        decoder_output = Reshape((128,), name="reshape_final")(x)
        decoder = Model(decoder_input, decoder_output, name="decoder")

        # Autoencoder
        encoded_spectrum = encoder(encoder_input)
        decoded_spectrum = decoder(encoded_spectrum)
        autoencoder = Model(encoder_input, decoded_spectrum, name="autoencoder")

        self.encoder = encoder
        self.decoder = decoder
        self.autoencoder = autoencoder

    def call(self, x):
        decoded = self.autoencoder(x)
        return decoded


class ARADv1(PyRIIDModel):
    """PyRIID-compatible ARAD v1 model supporting `SampleSet`s.
    """
    def __init__(self, model: ARADv1TF = None):
        """
        Args:
            model: a previously initialized TF implementation of ARADv1
        """
        super().__init__()

        self.model = model

        self._update_custom_objects("ARADv1TF", ARADv1TF)

    def fit(self, ss: SampleSet, epochs: int = 300, validation_split=0.2,
            es_verbose: int = 0, verbose: bool = False):
        """Fit a model to the given `SampleSet`.

        Args:
            ss: `SampleSet` of `n` spectra where `n` >= 1
            epochs: maximum number of training epochs
            validation_split: percentage of the training data to use as validation data
            es_verbose: verbosity level for `tf.keras.callbacks.EarlyStopping`
            verbose: whether to show detailed model training output

        Returns:
            reconstructed_spectra: output of ARAD model
        """
        _check_spectra(ss)

        x = ss.get_samples().astype(float)

        optimizer = tf.keras.optimizers.Nadam(
            learning_rate=1e-4
        )

        if not self.model:
            self.model = ARADv1TF()
        self.model.compile(
            loss=None,
            optimizer=optimizer
        )

        callbacks = [
            EarlyStopping(
                monitor="val_loss",
                patience=5,
                verbose=es_verbose,
                restore_best_weights=True,
                mode="min",
                min_delta=1e-7
            ),
            ReduceLROnPlateau(
                monitor="val_loss",
                factor=0.1,
                patience=3,
                min_delta=1e-8
            )
        ]

        history = self.model.fit(
            x=x,
            y=None,
            epochs=epochs,
            verbose=verbose,
            validation_split=validation_split,
            callbacks=callbacks,
            shuffle=True,
            batch_size=64
        )

        self._update_info(
            normalization=ss.spectra_state,
        )

        return history

    def predict(self, ss: SampleSet, verbose=False):
        """Generate reconstructions for given `SampleSet`.

        Args:
            ss: `SampleSet` of `n` spectra where `n` >= 1

        Returns:
            reconstructed_spectra: output of ARAD model
        """
        _check_spectra(ss)

        x = ss.get_samples().astype(float)

        reconstructed_spectra = self.model.predict(x, verbose=verbose)
        reconstruction_errors = entropy(x, reconstructed_spectra, axis=1)
        ss.info["recon_error"] = reconstruction_errors

        return reconstructed_spectra


class ARADv2(PyRIIDModel):
    """PyRIID-compatible ARAD v2 model supporting `SampleSet`s.
    """
    def __init__(self, model: ARADv2TF = None):
        """
        Args:
            model: a previously initialized TF implementation of ARADv1
        """
        super().__init__()

        self.model = model

        self._update_custom_objects("ARADv2TF", ARADv2TF)

    def fit(self, ss: SampleSet, epochs: int = 300, validation_split=0.2,
            es_verbose: int = 0, verbose: bool = False):
        """Fit a model to the given `SampleSet`.

        Args:
            ss: `SampleSet` of `n` spectra where `n` >= 1
            epochs: maximum number of training epochs
            validation_split: percentage of the training data to use as validation data
            es_verbose: verbosity level for `tf.keras.callbacks.EarlyStopping`
            verbose: whether to show detailed model training output

        Returns:
            reconstructed_spectra: output of ARAD model
        """
        _check_spectra(ss)

        x = ss.get_samples().astype(float)

        optimizer = tf.keras.optimizers.Adam(
            learning_rate=0.01,
            epsilon=0.05
        )

        if not self.model:
            self.model = ARADv2TF()
        self.model.compile(
            loss=jensen_shannon_distance,
            optimizer=optimizer
        )

        callbacks = [
            EarlyStopping(
                monitor="val_loss",
                patience=6,
                verbose=es_verbose,
                restore_best_weights=True,
                mode="min",
                min_delta=1e-4
            ),
            ReduceLROnPlateau(
                monitor="val_loss",
                factor=0.1,
                patience=3,
                min_delta=1e-4
            )
        ]

        history = self.model.fit(
            x=x,
            y=x,
            epochs=epochs,
            verbose=verbose,
            validation_split=validation_split,
            callbacks=callbacks,
            shuffle=True,
            batch_size=32
        )

        self._update_info(
            normalization=ss.spectra_state,
        )

        return history

    def predict(self, ss: SampleSet, verbose=False):
        """Generate reconstructions for given `SampleSet`.

        Args:
            ss: `SampleSet` of `n` spectra where `n` >= 1

        Returns:
            reconstructed_spectra: output of ARAD model
        """
        _check_spectra(ss)

        x = ss.get_samples().astype(float)

        reconstructed_spectra = self.model.predict(x, verbose=verbose)
        reconstruction_errors = jensenshannon(x, reconstructed_spectra, axis=1)
        ss.info["recon_error"] = reconstruction_errors

        return reconstructed_spectra


class ARADLatentPredictor(PyRIIDModel):
    """PyRIID-compatible model for branching from the latent space of a pre-trained
    ARAD model for a separate, arbitrary prediction task.
    """
    def __init__(self, hidden_layers: tuple = (8, 4,),
                 hidden_activation: str = "relu", final_activation: str = "linear",
                 loss: str = "mse", optimizer="adam", optimizer_kwargs=None,
                 learning_rate: float = 1e-3, metrics: tuple = ("mse", ),
                 kernel_l1_regularization: float = 0.0, kernel_l2_regularization: float = 0.0,
                 bias_l1_regularization: float = 0.0, bias_l2_regularization: float = 0.0,
                 activity_l1_regularization: float = 0.0, activity_l2_regularization: float = 0.0,
                 dropout: float = 0.0, **base_kwargs):
        """
        Args:
            hidden_layers: tuple defining the number and size of dense layers
            hidden_activation: activation function to use for each dense layer
            final_activation: activation function to use for final layer
            loss: loss function to use for training
            optimizer: tensorflow optimizer or optimizer name to use for training
            optimizer_kwargs: kwargs for optimizer
            learning_rate: optional learning rate for the optimizer
            metrics: list of metrics to be evaluating during training
            kernel_l1_regularization: l1 regularization value for the kernel regularizer
            kernel_l2_regularization: l2 regularization value for the kernel regularizer
            bias_l1_regularization: l1 regularization value for the bias regularizer
            bias_l2_regularization: l2 regularization value for the bias regularizer
            activity_l1_regularization: l1 regularization value for the activity regularizer
            activity_l2_regularization: l2 regularization value for the activity regularizer
            dropout: amount of dropout to apply to each dense layer
        """
        super().__init__(**base_kwargs)

        self.hidden_layers = hidden_layers
        self.hidden_activation = hidden_activation
        self.final_activation = final_activation
        self.loss = loss
        self.optimizer = optimizer
        if isinstance(optimizer, str):
            self.optimizer = tf.keras.optimizers.get(optimizer)
        if optimizer_kwargs is not None:
            for key, value in optimizer_kwargs.items():
                setattr(self.optimizer, key, value)
        self.optimizer.learning_rate = learning_rate
        self.metrics = metrics
        self.kernel_l1_regularization = kernel_l1_regularization
        self.kernel_l2_regularization = kernel_l2_regularization
        self.bias_l1_regularization = bias_l1_regularization
        self.bias_l2_regularization = bias_l2_regularization
        self.activity_l1_regularization = activity_l1_regularization
        self.activity_l2_regularization = activity_l2_regularization
        self.dropout = dropout
        self.model = None
        self.encoder = None

    def _initialize_model(self, arad: Model, output_size: int):
        """Build Keras MLP model.
        """
        encoder = arad.get_layer("encoder")
        encoder_input = encoder.get_layer(index=0).input
        encoder_output = encoder.get_layer(index=-1).output
        encoder_output_shape = encoder_output.shape

        predictor_input = Input(shape=encoder_output_shape, name="predictor_input")
        x = predictor_input
        for layer, nodes in enumerate(self.hidden_layers):
            x = tf.keras.layers.Dense(
                nodes,
                activation=self.hidden_activation,
                kernel_regularizer=L1L2(
                    l1=self.kernel_l1_regularization,
                    l2=self.kernel_l2_regularization
                ),
                bias_regularizer=L1L2(
                    l1=self.bias_l1_regularization,
                    l2=self.bias_l2_regularization
                ),
                activity_regularizer=L1L2(
                    l1=self.activity_l1_regularization,
                    l2=self.activity_l2_regularization
                ),
                name=f"dense_{layer}"
            )(x)
            if self.dropout > 0:
                x = tf.keras.layers.Dropout(self.dropout)(x)
        predictor_output = tf.keras.layers.Dense(
            output_size,
            activation=self.final_activation,
            name="output"
        )(x)
        predictor = Model(predictor_input, predictor_output, name="predictor")

        encoded_spectrum = encoder(encoder_input)
        predictions = predictor(encoded_spectrum)
        self.model = Model(encoder_input, predictions, name="predictor")
        # Freeze the layers corresponding to the autoencoder
        # Note: setting trainable to False is recursive to sub-layers per TF docs:
        # https://www.tensorflow.org/guide/keras/transfer_learning#recursive_setting_of_the_trainable_attribute
        for layer in self.model.layers[:-1]:
            layer.trainable = False

    def _check_targets(self, target_info_columns, target_level):
        """Check that valid target options are provided."""
        if target_info_columns and target_level:
            raise ValueError((
                "You have specified both target_info_columns (regression task) and "
                "a target_level (classification task), but only one can be set."
            ))
        if not target_info_columns and not target_level:
            raise ValueError((
                "You must specify either target_info_columns (regression task) or "
                "a target_level (classification task)."
            ))

    def fit(self, arad: Model, ss: SampleSet, target_info_columns: List[str] = None,
            target_level: str = None, batch_size: int = 10, epochs: int = 20,
            validation_split: float = 0.2, callbacks=None, patience: int = 15,
            es_monitor: str = "val_loss", es_mode: str = "min", es_verbose=0,
            es_min_delta: float = 0.0, verbose: bool = False):
        """Fit a model to the given SampleSet(s).

        Args:
            arad: a pretrained ARAD model (a TensorFlow Model object, not a PyRIIDModel wrapper)
            ss: `SampleSet` of `n` spectra where `n` >= 1
            target_info_columns: list of columns names from SampleSet info dataframe which
                denote what values the model should target
            target_level: `SampleSet.sources` column level to target for classification
            batch_size: number of samples per gradient update
            epochs: maximum number of training iterations
            validation_split: proportion of training data to use as validation data
            callbacks: list of callbacks to be passed to TensorFlow Model.fit() method
            patience: number of epochs to wait for tf.keras.callbacks.EarlyStopping object
            es_monitor: quantity to be monitored for tf.keras.callbacks.EarlyStopping object
            es_mode: mode for tf.keras.callbacks.EarlyStopping object
            es_verbose: verbosity level for tf.keras.callbacks.EarlyStopping object
            es_min_delta: minimum change to count as an improvement for early stopping
            verbose: whether model training output is printed to the terminal
        """
        self._check_targets(target_info_columns, target_level)

        x_train = ss.get_samples().astype(float)
        if target_info_columns:
            y_train = ss.info[target_info_columns].values.astype(float)
        else:
            source_contributions_df = ss.sources.groupby(
                axis=1,
                level=target_level,
                sort=False
            ).sum()
            y_train = source_contributions_df.values.astype(float)

        if not self.model:
            self._initialize_model(arad=arad, output_size=y_train.shape[1])

        self.model.compile(
            loss=self.loss,
            optimizer=self.optimizer,
            metrics=self.metrics
        )
        es = EarlyStopping(
            monitor=es_monitor,
            patience=patience,
            verbose=es_verbose,
            restore_best_weights=True,
            mode=es_mode,
            min_delta=es_min_delta
        )
        if callbacks:
            callbacks.append(es)
        else:
            callbacks = [es]
        history = self.model.fit(
            x_train,
            y_train,
            epochs=epochs,
            verbose=verbose,
            validation_split=validation_split,
            callbacks=callbacks,
            shuffle=True,
            batch_size=batch_size
        )

        self._update_info(
            normalization=ss.spectra_state,
            target_level=target_level,
            model_outputs=target_info_columns,
        )
        if target_level:
            self._update_info(
                model_outputs=source_contributions_df.columns.values.tolist(),
            )

        return history

    def predict(self, ss: SampleSet, verbose=False):
        spectra = ss.get_samples().astype(float)
        predictions = self.model.predict(spectra, verbose=verbose)

        if self.target_level:
            col_level_idx = SampleSet.SOURCES_MULTI_INDEX_NAMES.index(self.target_level)
            col_level_subset = SampleSet.SOURCES_MULTI_INDEX_NAMES[:col_level_idx+1]
            ss.prediction_probas = pd.DataFrame(
                data=predictions,
                columns=pd.MultiIndex.from_tuples(
                    self.get_model_outputs_as_label_tuples(),
                    names=col_level_subset
                )
            )

        ss.classified_by = self.model_id

        return predictions


def _check_spectra(ss: SampleSet):
    """Checks if SampleSet spectra are compatible with ARAD models."""
    if ss.n_samples <= 0:
        raise ValueError("No spectr[a|um] provided!")
    if not ss.all_spectra_sum_to_one():
        raise ValueError("All spectra must sum to one.")
    if not ss.spectra_state == SpectraState.L1Normalized:
        raise ValueError(
            f"SpectraState must be L1Normalzied, provided SpectraState is {ss.spectra_state}."
        )
    if not ss.n_channels == 128:
        raise ValueError(
            f"Spectra must have 128 channels, provided spectra have {ss.n_channels} channels."
        )



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module provides tools for handling data related to the
"Detecting Radiological Threats in Urban Areas" TopCoder Challenge.
https://doi.org/10.1038/s41597-020-00672-2
"""
import csv
import logging
import os
from pathlib import Path

import numpy as np
import pandas as pd

from riid import SAMPLESET_HDF_FILE_EXTENSION
from riid.data.converters import _validate_and_create_output_dir
from riid.data.labeling import label_to_index_element
from riid.data.sampleset import SampleSet

SOURCE_ID_TO_LABEL = {
    0: "Background",
    1: "HEU",
    2: "WGPu",
    3: "I131",
    4: "Co60",
    5: "Tc99m",
    6: "HEU + Tc99m",
}
DISTINCT_SOURCES = list(SOURCE_ID_TO_LABEL.values())


def _get_answers(answer_file_path: str):
    answers = {}
    with open(answer_file_path) as csvfile:
        reader = csv.reader(csvfile, delimiter=",")
        _ = next(reader)  # skip header
        # timestamp = 0  # in milliseconds
        for row in reader:
            run_id = row[0]
            source_id = int(row[1])
            source_time_secs = float(row[2])
            answers[run_id] = {
                "source_id": source_id,
                "source_time_secs": source_time_secs
            }
    return answers


def topcoder_file_to_ss(file_path: str, sample_interval: float, n_bins: int = 1024,
                        max_energy_kev: int = 4000, answers_path: str = None) -> SampleSet:
    """Convert a TopCoder CSV file of list-mode data into a SampleSet.

    Args:
        file_path: file path of the CSV file
        sample_interval: integration time (referred to as "live time" later) to use in seconds.
            Warning: the final sample in the set will likely be truncated, i.e., the count rate
            will appear low because the live time represented is too large.
            Consider ignoring the last sample.
        n_bins: desired number of bins in the resulting spectra.
            Bins will be uniformly spaced from 0 to `max_energy_kev`.
        max_energy_kev: desired maximum of the energy range represented in the resulting spectra.
            Intuition (assuming a fixed number of bins): a higher max energy value "compresses" the
            spectral information; a lower max energy value spreads out the spectral information and
            counts are potentially lost off the high energy end of the specturm.
        answers_path: path to the answer key for the data. If provided, this will fill out the
            `SampleSet.sources` DataFrame.

    Returns:
        `SampleSet` containing the series of spectra for a single run
    """
    file_name_with_dir = os.path.splitext(file_path)[0]
    file_name = os.path.basename(file_name_with_dir)
    slice_duration_ms = sample_interval * 1000  # in milliseconds
    events = []
    with open(file_path) as csvfile:
        reader = csv.reader(csvfile, delimiter=",")
        timestamp = 0  # in milliseconds
        for row in reader:
            timestamp += int(row[0]) / 1000  # microseconds to milliseconds
            energy = float(row[1])
            if energy > max_energy_kev:
                msg = (
                    f"Encountered energy ({energy:.2f} keV) greater than "
                    f"specified max energy ({max_energy_kev:.2f})"
                )
                logging.warn(msg)
            channel = int(n_bins * energy // max_energy_kev)  # energy to bin
            events.append((timestamp, channel, 1))

    events_df = pd.DataFrame(
        events,
        columns=["timestamp", "channel", "counts"]
    )
    # Organize events into time intervals
    event_time_intervals = pd.cut(
        events_df["timestamp"],
        np.arange(
            start=0,
            stop=events_df["timestamp"].max()+slice_duration_ms,
            step=slice_duration_ms
        )
    )
    # Group events by time intervals
    event_time_groups = events_df.groupby(event_time_intervals)
    # Within time intervals, sum counts by channel
    result = event_time_groups.apply(
        lambda x: x.groupby("channel").sum().loc[:, ["counts"]]
    )
    # Create new dataframe where: row = time interval, column = channel
    spectra_df = result.unstack(level=-1, fill_value=0)
    spectra_df = spectra_df["counts"]
    # Add in missing columns as needed
    col_list = np.arange(0, n_bins)
    cols_to_add = np.setdiff1d(col_list, spectra_df.columns)
    missing_df = pd.DataFrame(
        0,
        columns=cols_to_add,
        index=spectra_df.index
    )
    combined_unsorted_df = pd.concat([spectra_df, missing_df], axis=1)
    combined_df = combined_unsorted_df.reindex(
        sorted(combined_unsorted_df.columns),
        axis=1
    )
    spectra_df = combined_df.astype(int)
    spectra_df.reset_index(inplace=True, drop=True)

    # SampleSet creation
    ss = SampleSet()
    ss.measured_or_synthetic = "synthetic"
    ss.detector_info = {
        "name": "2\"x4\"x16\" NaI(Tl)",
        "height_cm": 100,
        "fwhm_at_661_kev": 0.075,
    }
    ss.spectra = spectra_df
    ss.info.total_counts = spectra_df.sum(axis=1)
    ss.info.live_time = sample_interval
    ss.info.description = file_name
    ss.info.ecal_order_0 = 0
    ss.info.ecal_order_1 = max_energy_kev
    ss.info.ecal_order_2 = 0
    ss.info.ecal_order_3 = 0
    ss.info.ecal_low_e = 0

    if answers_path:
        answers = _get_answers(answers_path)
        run_id = file_name.split("runID-")[-1].split(".")[0]
        ss.info.timestamp = answers[run_id]["source_time_secs"]
        source_id = answers[run_id]["source_id"]
        sources_mi = pd.MultiIndex.from_tuples([
            label_to_index_element(x, label_level="Seed")
            for x in DISTINCT_SOURCES
        ], names=SampleSet.SOURCES_MULTI_INDEX_NAMES)
        sources_df = pd.DataFrame(
            np.zeros((ss.n_samples, len(DISTINCT_SOURCES))),
            columns=sources_mi,
        )
        sources_df.sort_index(axis=1, inplace=True)
        source = label_to_index_element(SOURCE_ID_TO_LABEL[source_id], label_level="Seed")
        sources_df[source] = 1.0
        ss.sources = sources_df

    return ss


def convert_and_save(input_file_path: str, output_dir: str = None,
                     skip_existing: bool = True, **kwargs):
    """Convert TopCoder file to SampleSet and save as HDF.

    Output file will have same name with different extension.

    Args:
        input_file_path: file path of the CSV file
        output_dir: alternative directory in which to save HDF files
            (defaults to `input_file_path` parent if not provided)
        skip_existing: whether to skip conversion if the file already exists
        kwargs: keyword args passed to `topcoder_file_to_ss()`
    """
    input_path = Path(input_file_path)
    if not output_dir:
        output_dir = input_path.parent
    _validate_and_create_output_dir(output_dir)
    output_file_path = os.path.join(output_dir, input_path.stem + SAMPLESET_HDF_FILE_EXTENSION)
    if skip_existing and os.path.exists(output_file_path):
        return

    ss = topcoder_file_to_ss(input_file_path, **kwargs)
    ss.to_hdf(output_file_path)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
"""This module contains the `SampleSet` class and other `SampleSet`-related functions."""
from __future__ import \
    annotations  # Enables SampleSet hints inside SampleSet itself

import copy
import json
import logging
import operator
import os
import random
import re
import warnings
from datetime import datetime
from enum import Enum
from typing import Callable, Iterable, Tuple, Union

import numpy as np
import pandas as pd
import tqdm
from scipy import stats
from scipy.interpolate import interp1d
from scipy.spatial import distance

import riid
from riid.data import get_expected_spectra
from riid.data.labeling import (NO_CATEGORY, NO_ISOTOPE, NO_SEED,
                                _find_category, _find_isotope)
from riid.gadras.pcf import (_dict_to_pcf, _pack_compressed_text_buffer,
                             _pcf_to_dict, _unpack_compressed_text_buffer)


class SpectraState(int, Enum):
    """States in which SampleSet spectra can exist."""
    Unknown = 0
    Counts = 1
    L1Normalized = 2
    L2Normalized = 3


class SpectraType(int, Enum):
    """Types for SampleSet spectra."""
    Unknown = 0
    Background = 1
    Foreground = 2
    Gross = 3
    BackgroundForeground = 4


class SampleSet():
    """A collection of spectrum samples and their metadata."""
    DEAD_TIME_PROP_INFO_KEY = "dead_time_prop"
    SOURCES_MULTI_INDEX_NAMES = (
        "Category",
        "Isotope",
        "Seed"
    )
    ECAL_INFO_COLUMNS = (
        "ecal_order_0",
        "ecal_order_1",
        "ecal_order_2",
        "ecal_order_3",
        "ecal_low_e"
    )
    DEFAULT_INFO_COLUMNS = (
        "description",
        "timestamp",
        "live_time",
        "real_time",
        "total_counts",
        "snr",
        "neutron_counts",
        "distance_cm",
        *ECAL_INFO_COLUMNS,
        "areal_density",
        "atomic_number",
        "occupancy_flag",
        "tag",
    )
    DEFAULT_BG_SEED_NAMES = (
        "Cosmic",
        "PotassiumInSoil",
        "UraniumInSoil",
        "ThoriumInSoil",
    )
    SUPPORTED_STATES_FOR_ARITHMETIC = (
        SpectraState.Counts.value,
        SpectraState.L1Normalized.value
    )
    DEFAULT_EXCLUSIONS_FROM_COMPARISON = (
        "timestamp",
        "description",
        "occupancy_flag",
        "tag",
    )

    def __init__(self):
        """
        Expected sizes for DataFrames:
            self._spectra: [n_samples, n_channels]
            self._sources: [n_samples, n_sources]
        """
        self._spectra = pd.DataFrame()
        self._sources = pd.DataFrame()
        self._info = pd.DataFrame(columns=SampleSet.DEFAULT_INFO_COLUMNS)
        self._detector_info = {}
        self._synthesis_info = {}
        self._prediction_probas = pd.DataFrame()
        self._measured_or_synthetic = None
        self.pyriid_version = riid.__version__
        self._spectra_state = SpectraState.Unknown
        self._spectra_type = SpectraType.Unknown
        self._classified_by = ""

    def __bool__(self):
        return bool(len(self))

    def __getitem__(self, key: Union[slice, int, list]):
        selection = key
        if isinstance(key, int):
            selection = slice(key, key+1)
        elif isinstance(key, pd.Series) and key.dtype == bool:
            selection = self.spectra.index[key]

        sub_ss = copy.copy(self)
        sub_ss.spectra = sub_ss.spectra.iloc[selection].reset_index(drop=True)
        sub_ss.sources = sub_ss.sources.iloc[selection].reset_index(drop=True)
        sub_ss.info = sub_ss.info.iloc[selection].reset_index(drop=True)
        if not sub_ss.prediction_probas.empty:
            sub_ss.prediction_probas = sub_ss.prediction_probas.iloc[selection] \
                .reset_index(drop=True)

        return sub_ss

    def __len__(self):
        return self.n_samples

    def __str__(self):
        UNKNOWN_VALUE = "Unknown"
        detector_info_str = self.detector_info if self.detector_info else UNKNOWN_VALUE
        sources_present_str = ", ".join(sorted(np.unique(self.get_labels()))) \
            if not self.sources.empty \
            else UNKNOWN_VALUE
        predictions_str = ", ".join(sorted(np.unique(self.get_labels()))) \
            if not self.prediction_probas.empty \
            else "None"
        pyriid_version_str = self.pyriid_version if self.pyriid_version else UNKNOWN_VALUE
        info_dict = {
            "# of samples": self.n_samples,
            "# of channels": self.n_channels,
            "Detector": detector_info_str,
            "Present predictions": predictions_str,
            "Present sources": sources_present_str,
            "PyRIID version": pyriid_version_str,
        }
        summary = "\n".join(_dict_to_bulleted_list(info_dict))
        return f"SampleSet Summary:\n{summary}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, ss: SampleSet):
        return np.allclose(self._spectra.values, ss._spectra.values, atol=1e-3)

    def _check_arithmetic_supported(self, ss2: SampleSet):
        if ss2.n_samples != 1 and ss2.n_samples != self.n_samples:
            n_sample_str = (
                "You can only add/subtract SampleSet objects "
                "when the second SampleSet has only one spectrum "
                "(the spectrum will be scaled to match each sample of the first SampleSet) or"
                "when the second SampleSet has the same number of samples as the first "
                "(no rescaling will occur)."
            )
            raise InvalidSampleCountError(n_sample_str)
        if self.n_channels != ss2.n_channels:
            channel_str = f"({self.n_channels} != {ss2.n_channels})"
            raise ChannelCountMismatchError(f"Mismatched spectra channels {channel_str}!")
        if self.spectra_state != ss2.spectra_state:
            state_str = f"({self.spectra_state} != {ss2.spectra_state})"
            raise SpectraStateMismatchError(f"Mismatched spectra states {state_str}!")
        if self.spectra_state.value not in self.SUPPORTED_STATES_FOR_ARITHMETIC:
            raise ValueError(f"{self.spectra_state} spectra state not supported for arithmetic!")
            # Don't need to check spectra state of ss2 since they passed the prior equality check

    def _get_scaled_spectra(self, ss: SampleSet) -> np.ndarray:
        spectra_in_counts = ss.spectra.iloc[0].values.copy()
        if ss.spectra_state == SpectraState.L1Normalized:
            spectra_in_counts *= ss.info.iloc[0].total_counts
        live_times = ss.info.iloc[0].live_time
        spectra_in_cps = spectra_in_counts / live_times
        scaled_spectra = np.concatenate(
            spectra_in_cps * self.info.live_time.values[:, np.newaxis, np.newaxis]
        )
        return scaled_spectra

    def _get_arithmetic_result(self, ss: SampleSet, op) -> SampleSet:
        self._check_arithmetic_supported(ss)

        if ss.n_samples == 1:
            scaled_spectra = self._get_scaled_spectra(ss)
        else:
            scaled_spectra = ss.spectra.values  # Assumed to already be scaled
        new_ss = self[:]

        is_l1_normalized = new_ss.spectra_state == SpectraState.L1Normalized
        if is_l1_normalized:
            new_ss.spectra = new_ss.spectra.multiply(new_ss.info.total_counts, axis=0)

        new_ss.spectra = op(new_ss.spectra, scaled_spectra)

        if is_l1_normalized:
            new_ss.normalize()

        return new_ss

    def __add__(self, bg_ss: SampleSet) -> SampleSet:
        """Add the given background spectr(um|a) to the spectra of the current SampleSet.
        """
        if bg_ss.spectra_type.value != SpectraType.Background.value:
            msg = (
                "`bg_ss` argument must have a `spectra_type` of `Background`."
                f"Its `spectra_type` is `{bg_ss.spectra_type}`."
            )
            raise ValueError(msg)
        if self.spectra_type.value != SpectraType.Gross.value and \
           self.spectra_type.value != SpectraType.Foreground.value:
            msg = (
                "Current `SampleSet` must have a `spectra_type` of `Gross` or `Foreground`. "
                f"Current `spectra_type` is `{self.spectra_type}`."
            )
            raise ValueError(msg)

        new_ss = self._get_arithmetic_result(bg_ss, operator.add)
        new_ss.spectra_type = SpectraType.Gross
        return new_ss

    def __sub__(self, ss: SampleSet) -> SampleSet:
        """Subtract the given background or foreground spectr(um|a) from the spectra of the current
        `SampleSet`.
        """
        if ss.spectra_type.value != SpectraType.Background.value and \
           ss.spectra_type.value != SpectraType.Foreground.value:
            msg = (
                "`ss` argument must have a `spectra_type` of `Background` of `Foreground`."
                f"Its `spectra_type` is `{ss.spectra_type}`."
            )
            raise ValueError(msg)
        if self.spectra_type.value != SpectraType.Gross.value:
            msg = (
                "Current `SampleSet` must have a `spectra_type` of `Gross`. "
                f"Current `spectra_type` is `{self.spectra_type}`."
            )
            raise ValueError(msg)

        new_ss = self._get_arithmetic_result(ss, operator.sub)
        if ss.spectra_type == SpectraType.Background:
            new_ss.spectra_type = SpectraType.Foreground
        else:
            new_ss.spectra_type = SpectraType.Background
        return new_ss

    # region Properties

    @property
    def category_names(self):
        """Get the names of the categories involved in this SampleSet."""
        return self.sources.columns.get_level_values("Category")

    @property
    def classified_by(self):
        """Get or set the UUID of the model that classified the SampleSet.

        TODO: Implement as third dimension to `prediction_probas` DataFrame
        """
        return self._classified_by

    @classified_by.setter
    def classified_by(self, value):
        self._classified_by = value

    @property
    def detector_info(self):
        """Get or set the detector info on which this `SampleSet` is based.

        TODO: implement as DataFrame
        """
        return self._detector_info

    @detector_info.setter
    def detector_info(self, value):
        self._detector_info = value

    @property
    def difficulty_score(self, mean=10.0, std=3.0) -> float:
        """Compute the "difficulty" of the `SampleSet` on a scale of 0 to 1,
        where 0 is easiest and 1 is hardest.

        The difficulty of a `SampleSet` is the mean of the individual sample difficulties.
        Each sample's difficulty is determined by where its signal strength (SNR)
        falls on the survival function (AKA reliability function, or 1 - CDF) for a
        logistic distribution.
        The decision to use this distribution is based on empirical data showing that
        classifier performance increases in a logistic fashion as signal strength increases.

        The primary use of this function is for situations where ground truth is unknown
        and you want to get an idea of how "difficult" it will be for a model
        to make predictions.  Consider the following example:
            Given a `SampleSet` for which ground truth is unknown and the signal strength for
            each spectrum is low.
            A model considering, say, 100+ isotopes will see this `SampleSet` as quite
            difficult, whereas a model considering 5 isotopes will, being more specialized,
            see the `SampleSet` as easier.
            Of course, this makes the assumption that both models were trained to the
            isotope(s) contained in the test `SampleSet`.

        Based on the previous example, the unfortunate state of this function is that you
        must know how to pick means and standard deviations which properly reflect the:

        - number of target isotopes
        - amount of variation within each isotope (i.e., shielding, scattering, etc.)
        - detector resolution

        A future goal of ours is to provide updates to this function and docstring which make
        choosing the mean and standard deviation easier/automated based on more of our findings.

        Arguments:
            mean: SNR value representing the mean of the logistic distribution
            std: standard deviation of the logistic distribution

        Returns:
            The mean of all SNR values passed through a logistic survival function
        """
        snrs: np.ndarray = self.info.snr.clip(1e-6)
        score = float(stats.logistic.sf(snrs, loc=mean, scale=std).mean())

        return score

    @property
    def ecal(self):
        """Get or set the ecal terms."""
        ecal_terms = self.info[list(self.ECAL_INFO_COLUMNS)].values
        return ecal_terms

    @ecal.setter
    def ecal(self, value):
        self.info.loc[:, self.ECAL_INFO_COLUMNS[0]] = value[0]
        self.info.loc[:, self.ECAL_INFO_COLUMNS[1]] = value[1]
        self.info.loc[:, self.ECAL_INFO_COLUMNS[2]] = value[2]
        self.info.loc[:, self.ECAL_INFO_COLUMNS[3]] = value[3]
        self.info.loc[:, self.ECAL_INFO_COLUMNS[4]] = value[4]

    @property
    def info(self):
        """Get or set the info DataFrame."""
        return self._info

    @info.setter
    def info(self, value):
        self._info = value

    @property
    def isotope_names(self):
        """Get the names of the isotopes involved in this SampleSet."""
        return self.sources.columns.get_level_values("Isotope")

    @property
    def measured_or_synthetic(self):
        """Get or set the value for measured_or_synthetic for the SampleSet object."""
        return self._measured_or_synthetic

    @measured_or_synthetic.setter
    def measured_or_synthetic(self, value):
        self._measured_or_synthetic = value

    @property
    def n_channels(self):
        """Get the number of channels included in the spectra DataFrame.
        Channels may also be referred to as bins.
        """
        return self._spectra.shape[1]

    @property
    def n_samples(self):
        """Get the number of samples included in the spectra dataframe,
        where each row is a sample.
        """
        return self._spectra.shape[0]

    @property
    def prediction_probas(self):
        """Get or set the prediction_probas DataFrame."""
        return self._prediction_probas

    @prediction_probas.setter
    def prediction_probas(self, value):
        if isinstance(value, pd.DataFrame):
            self._prediction_probas = value
        else:
            self._prediction_probas = pd.DataFrame(value, columns=self.sources.columns)

    @property
    def seed_names(self):
        """Get the names of the seeds involved in this `SampleSet`."""
        return self.sources.columns.get_level_values("Seed")

    @property
    def sources(self):
        """Get or set the sources `DataFrame`."""
        return self._sources

    @sources.setter
    def sources(self, value):
        self._sources = value.replace(np.nan, 0)

    @property
    def spectra(self):
        """Get or set the spectra `DataFrame`."""
        return self._spectra

    @spectra.setter
    def spectra(self, value):
        self._spectra = value

    @property
    def spectra_state(self: SpectraState):
        """Get or set the spectra state."""
        return self._spectra_state

    @spectra_state.setter
    def spectra_state(self, value: SpectraState):
        self._spectra_state = value

    @property
    def spectra_type(self: SpectraType):
        """Get or set the spectra type."""
        return self._spectra_type

    @spectra_type.setter
    def spectra_type(self, value: SpectraType):
        self._spectra_type = value

    @property
    def synthesis_info(self):
        """Get or set the information that was used to by a synthesizer."""
        return self._synthesis_info

    @synthesis_info.setter
    def synthesis_info(self, value):
        self._synthesis_info = value

    # endregion

    def _channels_to_energies(self, fractional_energy_bins,
                              offset: float, gain: float, quadratic: float, cubic: float,
                              low_energy: float) -> np.ndarray:
        r"""Convert fractional energy bins to the energy represented by the lower edge of the bin.

        A spectrum represents a range of energy (e.g., 0 keV to 3000 keV).
        The energy range represented by a spectrum can be inferred from its energy calibration
        (e-cal for short), if known.  Using the convention in ANSI N42.42-2006, the e-cal is
        defined using the following terms:

        - \( a_0 \): offset (ecal_order_0)
        - \( a_1 \): gain (ecal_order_1)
        - \( a_2 \): quadratic (ecal_order_2)
        - \( a_3 \): cubic (ecal_order_3)
        - \( a_4 \): low energy (ecal_low_e)

        The energy represented by the lower edge of channel \( i \) is then calculated as follows:

        $$
        E_i = a_0 + a_1 * x + a_2 * x^2 + a_3 * a^3 + \frac{a_4}{1 + 60x}
        $$

        where \( x \) is the fractional energy of the spectral range.
        Also, under this scheme and ignoring the non-linear terms (e.g., \( a_2 \) through
        \( a_4 \)), the range of energy represented by a spectrum can be quickly inferred
        from the e-cal as:

        - `min_energy = offset`
        - `max_energy = gain + offset`

        Note: spline interpolation is another way to characterize the energy calibration of a
        spectrum. In GADRAS, this is accomplished via deviation pairs; PyRIID does not currently
        support this but may in the future.

        Args:
            fractional_energy_bins: array of fractions representing the spacing of the channels
            offset: e-cal order 0 term
            gain: e-cal order 1 term
            quadratic: e-cal order 2 term
            cubic: e-cal order 3 term
            low_energy: e-cal low energy term

        Returns:
            Array containing the energies represented by the lower edge of each channel
            (hape depends on the shape of input values)
        """
        channel_energies = \
            offset + \
            fractional_energy_bins * gain + \
            fractional_energy_bins**2 * quadratic + \
            fractional_energy_bins**3 * cubic + \
            low_energy / (1 + 60 * fractional_energy_bins)
        return channel_energies

    def _check_target_level(self, target_level,
                            levels_allowed=SOURCES_MULTI_INDEX_NAMES):
        if target_level not in levels_allowed:
            raise ValueError((
                f"'{target_level}' is not an appropriate target level. "
                f"Acceptable values are: {levels_allowed}."
            ))

    def _get_dead_time_proportions(self):
        live_times = self.info.live_time.values
        real_times = self.info.real_time.values
        dead_time_props = (1.0 - live_times / real_times)
        dead_time_props = np.nan_to_num(dead_time_props, nan=1.0)
        return dead_time_props

    def all_spectra_sum_to_one(self, rtol: float = 0.0, atol: float = 1e-4) -> bool:
        """Checks if all spectra are normalized to sum to one."""
        spectra_counts = self.spectra.sum(axis=1).values
        all_sum_to_one = np.all(np.isclose(spectra_counts, 1, rtol=rtol, atol=atol))
        return all_sum_to_one

    def as_ecal(self, new_offset: float, new_gain: float,
                new_quadratic: float, new_cubic: float,
                new_low_energy: float) -> SampleSet:
        """Re-bin spectra based on energy by interpolating the current shape from the current
        binning structure to a new one.

        Warning: this operation is likely to add or remove spectral information
        depending on the new energy calibration values used.

        Args:
            new_offset: new gain value, i.e. the 0-th e-cal term
            new_gain: new gain value, i.e. the 1-th e-cal term
            new_quadratic: new quadratic value, i.e. the 2-th e-cal term
            new_cubic: new cubic value, i.e. the 3-th e-cal term
            new_low_energy: new low energy term

        Raises:
            `ValueError` when no argument values are provided
        """
        new_args = [new_offset, new_gain, new_quadratic, new_cubic, new_low_energy]
        if all(v is None for v in new_args):
            raise ValueError("At least one argument value must be provided.")

        ecal_cols = list(self.ECAL_INFO_COLUMNS)
        new_ecal = self.info[ecal_cols].copy()
        new_ecal.ecal_order_0 = new_offset
        new_ecal.ecal_order_1 = new_gain
        new_ecal.ecal_order_2 = new_quadratic
        new_ecal.ecal_order_3 = new_cubic
        new_ecal.ecal_low_e = new_low_energy

        all_original_channel_energies = self.get_all_channel_energies()
        all_new_channel_energies = []
        fractional_energy_bins = np.linspace(0, 1, self.n_channels)
        for i in range(self.n_samples):
            offset, gain, quadratic, cubic, low_energy = new_ecal.iloc[i].values
            new_channel_energies = self._channels_to_energies(
                fractional_energy_bins,
                offset, gain, quadratic, cubic, low_energy
            )
            all_new_channel_energies.append(new_channel_energies)

        new_spectra = np.zeros([self.n_samples, self.n_channels])
        # Perform interpolation
        for i in range(self.n_samples):
            f = interp1d(
                all_original_channel_energies[i],
                self.spectra.values[i],
                kind="slinear",
                fill_value=0,
                bounds_error=False
            )
            new_spectrum = f(all_new_channel_energies[i])
            new_spectra[i] = new_spectrum

        # Update spectra and info DataFrames
        new_ss = self[:]
        new_ss.spectra = pd.DataFrame(new_spectra)
        new_ss.info.total_counts = new_ss.spectra.sum(axis=1)
        new_ss.info[ecal_cols] = new_ecal
        return new_ss

    def check_seed_health(self, dead_time_threshold=1.0):
        """Checks health of all spectra and info assuming they are seeds.

        Invalidate states for which we currently check:

        - spectra do not sum to 1
        - dead time greater than or equal to provided threshold

        Args:
            dead_time_threshold: value at which seed dead time is unacceptable

        Raises:
            `AssertionError` if any check fails
        """
        all_spectra_sum_to_one = self.all_spectra_sum_to_one()
        assert all_spectra_sum_to_one

        if self.DEAD_TIME_PROP_INFO_KEY not in self.info.columns:
            self.set_dead_time_proportions()
        dead_samples = self.info[self.DEAD_TIME_PROP_INFO_KEY] >= dead_time_threshold
        all_samples_are_alive = not np.any(dead_samples)
        assert all_samples_are_alive

    def clip_negatives(self, min_value: float = 0):
        """Clip spectrum values to some minimum value.

        Args:
            min_value: value to which to clip existing spectrum values
        """
        self._spectra = pd.DataFrame(data=self._spectra.clip(min_value))

    def compare_to(self, ss: SampleSet, n_bins: int = 20, density: bool = False) \
            -> Tuple[dict, dict, dict]:
        """Compare the current `SampleSet` to another `SampleSet`.

        The function only compares the selected, mutual information of
        each `SampleSet` by computing the Jensen-Shannon distance between
        histograms of that information.

        Args:
            ss: `SampleSet` to compare to
            n_bins: number of bins we will sort both sample sets by
            density: whether histograms should be in counts or density

        Returns:
            Tuple of the following:

            - ss1_stats: dictionary of stats describing the first `SampleSet`
            - ss2_stats: dictionary of stats describing the second `SampleSet`
            - col_comparisons: dictionary of distance values comparing each stat
        """
        TARGET_SUMMARY_STATS = ["min", "max", "median", "mean", "std"]
        STAT_PRECISION = 3

        # Get info columns we want to compare
        info_columns = [x for x in list(self.DEFAULT_INFO_COLUMNS)
                        if x not in list(self.DEFAULT_EXCLUSIONS_FROM_COMPARISON)]

        # Get both sample sets comparable columns of data
        info_df1 = self.info[info_columns]
        info_df2 = ss.info[info_columns]

        # Check valid columns in each sample set (cannot have None or 0)
        ss1_valid_cols = [
            c for c in info_df1.columns
            if pd.to_numeric(info_df1[c], errors="coerce").notnull().all() and any(info_df1[c])
        ]
        ss2_valid_cols = [
            c for c in info_df2.columns
            if pd.to_numeric(info_df2[c], errors="coerce").notnull().all() and any(info_df2[c])
        ]

        # Remove non shared column lists
        for i in ss1_valid_cols:
            if i not in ss2_valid_cols:
                ss1_valid_cols.remove(i)
        for i in ss2_valid_cols:
            if i not in ss1_valid_cols:
                ss2_valid_cols.remove(i)

        # Remove non shared column data
        info_df1 = info_df1[ss1_valid_cols]
        info_df2 = info_df2[ss2_valid_cols]

        # Bin the data
        ss1_stats = {}
        ss2_stats = {}
        col_comparisons = {}
        for i in ss1_valid_cols:
            ss1_stats[i] = {}
            ss2_stats[i] = {}
            hist_range = (
                min(min(info_df1[i]), min(info_df2[i])),
                max(max(info_df1[i]), max(info_df2[i]))
            )
            # Get data in bins and get counts for each bin
            hist1, bins1 = np.histogram(info_df1[i], bins=n_bins, range=hist_range, density=density)
            hist2, bins2 = np.histogram(info_df2[i], bins=n_bins, range=hist_range, density=density)
            ss1_stats[i]["density"] = density
            ss2_stats[i]["density"] = density
            ss1_stats[i]["hist"] = hist1
            ss2_stats[i]["hist"] = hist2
            ss1_stats[i]["bins"] = bins1
            ss2_stats[i]["bins"] = bins2
            ss1_stats[i]["range"] = hist_range
            ss2_stats[i]["range"] = hist_range

            summary_stats_df1 = info_df1[i].agg(TARGET_SUMMARY_STATS).round(decimals=STAT_PRECISION)
            ss1_stats[i].update(summary_stats_df1.to_dict())
            summary_stats_df2 = info_df2[i].agg(TARGET_SUMMARY_STATS).round(decimals=STAT_PRECISION)
            ss2_stats[i].update(summary_stats_df2.to_dict())

            js_dist = distance.jensenshannon(hist1, hist2, axis=0)
            col_comparisons[i] = js_dist

        return ss1_stats, ss2_stats, col_comparisons

    def concat(self, ss_list: list):
        """Vertically concatenate multiple `SampleSet`s into one `SampleSet`.

        Combining of non-DataFrame properties (e.g., detector_info, synthesis_info, etc.)
        is NOT performed - it is the responsiblity of the user to ensure proper bookkeeping for
        these properties when concatenating `SampleSet`s.

        This method is most applicable to `SampleSet`s containing measured data from the same
        detector which could not be made as a single `SampleSet` because of how measurements had to
        be taken (i.e., measurements were taken by distinct processes separated by time).
        Therefore, the user should avoid concatenating `SampleSet`s when:

        - the data is from different detectors
        (note that `SampleSet`s are given to models for prediction, and models should be 1-to-1
        with physical detectors);
        - a mix of synthetic and measured data would occur (this could make an existing
        `synthesis_info` value ambiguous, but one way to handle this would be to add a new column
        to `info` distinguishing between synthetic and measured on a per-sample basis).

        Args:
            ss_list: list of `SampleSet`s to concatenate

        Returns:
            `SampleSet` object
        """
        if not ss_list:
            return
        if not isinstance(ss_list, list) and not isinstance(ss_list, tuple):
            ss_list = [ss_list]
        spectra_types = list(
            set([self.spectra_type.value] + [x.spectra_type.value for x in ss_list])
        )
        n_spectra_types = len(spectra_types)
        if n_spectra_types == 2:
            if SpectraType.Background.value in spectra_types and \
               SpectraType.Foreground.value in spectra_types:
                self.spectra_type = SpectraType.BackgroundForeground
            else:
                self.spectra_type = SpectraType.Unknown
        elif n_spectra_types == 0 or n_spectra_types > 2:
            self.spectra_type = SpectraType.Unknown

        self._spectra = pd.concat(
            [self._spectra] + [ss.spectra for ss in ss_list],
            ignore_index=True,
            sort=False
        )
        self._sources = pd.concat(
            [self._sources] + [ss.sources for ss in ss_list],
            ignore_index=True,
            sort=False
        )
        self._sources = self._sources.where(pd.notnull(self._sources), 0)
        self._info = pd.concat(
            [self._info] + [ss.info for ss in ss_list],
            ignore_index=True,
            sort=False
        )
        self._info = self._info.where(pd.notnull(self._info), None)
        self._prediction_probas = pd.concat(
            [self._prediction_probas] + [ss.prediction_probas for ss in ss_list],
            ignore_index=True,
            sort=False
        )
        self._prediction_probas = self._prediction_probas.where(
            pd.notnull(self._prediction_probas),
            0
        )

    def downsample_spectra(self, target_bins: int = 128, min_frac=1e-8):
        """Uniformly down-bin spectra.

        Args:
            target_bins: number of channels to which to bin the existing spectra

        Raises:
            `ValueError` when binning is not a multiple of the target binning
        """
        if self.n_channels % target_bins == 0:
            n_per_group = int(self.n_channels / target_bins)
            transformation = np.zeros([target_bins, self.n_channels])
            for i in range(target_bins):
                transformation[i, (i * n_per_group):((i * n_per_group) + n_per_group)] = 1
        else:
            msg = "Current binning ({}) is not a multiple of the target binning ({})".format(
                self.n_channels,
                target_bins
            )
            raise ValueError(msg)
        self._spectra = pd.DataFrame(
            data=np.matmul(
                self._spectra.values,
                transformation.T
            )
        )
        self._spectra[self._spectra < min_frac] = 0

    def drop_sources(self, col_names: Iterable = DEFAULT_BG_SEED_NAMES,
                     normalize_sources: bool = True,
                     target_level: str = "Seed"):
        """Drop columns from `sources`.

        Args:
            col_names: names of the sources columns to be dropped.
                The names of background seeds are used by default as removing the
                ground truth for background sources when dealing with synthetic,
                gross spectra is a common operation.
            normalize_sources: whether to normalize the sources DataFrame after dropping
                the column(s)
            target_level: `SampleSet.sources` column level to use
        """
        self.sources.drop(col_names, axis=1, inplace=True, level=target_level)
        if normalize_sources:
            self.normalize_sources()

    def drop_sources_columns_with_all_zeros(self):
        """Remove columns from the sources `DataFrame` that contain only zeros.

        Modifications are made in-place.
        """
        idxs = (self.sources != 0).any(axis=0)
        self.sources = self.sources.loc[:, idxs]
        self.sources.columns = self.sources.columns.remove_unused_levels()

        return idxs

    def drop_spectra_with_no_contributors(self) -> np.ndarray:
        """Remove samples where the spectrum has no recorded contributor.

        Modifications are inplace.

        Returns:
            A boolean mask of which rows were kept
        """
        idxs = np.where(self.sources.values.sum(axis=1) == 0)[0]
        kept_mask = ~np.in1d(self.sources.index.values, idxs)

        self.spectra = self.spectra.drop(index=idxs).reset_index(drop=True)
        self.sources = self.sources.drop(index=idxs).reset_index(drop=True)
        self.info = self.info.drop(index=idxs).reset_index(drop=True)

        return kept_mask

    def extend(self, spectra: Union[dict, list, np.array, pd.DataFrame],
               sources: pd.DataFrame, info: Union[list, pd.DataFrame]):
        """Extend the current SampleSet with the given data.

        Always verify that the data was appended in the way you expect based on what input type
        you are preferring to work with.

        Args:
            spectra: spectra to append to the current spectra DataFrame
            sources: sources to append to the current sources DataFrame
            info: info to append to the current info DataFrame

        Raises:
            `ValueError` when internal `DataFrame` lengths do not match
        """
        if not spectra.shape[0] == sources.shape[0] == info.shape[0]:
            msg = "The number of rows in each of the required positional arguments must be same."
            raise ValueError(msg)

        self._spectra = self._spectra\
            .append(pd.DataFrame(spectra), ignore_index=True, sort=True)\
            .fillna(0)

        new_sources_column_index = pd.MultiIndex.from_tuples(
            sources.keys(),
            names=SampleSet.SOURCES_MULTI_INDEX_NAMES
        )
        new_sources_df = pd.DataFrame(sources.values(), columns=new_sources_column_index)
        self._sources = self._sources\
            .append(new_sources_df, ignore_index=True, sort=True)\
            .fillna(0)

        self._info = self._info\
            .append(pd.DataFrame(info), ignore_index=True, sort=True)

    def get_all_channel_energies(self, fractional_energy_bins=None) -> np.ndarray:
        """Get the energy (in keV) represented by the lower edge of each channel for all samples.

        See docstring for `_channels_to_energies()` for more details.

        Args:
            fractional_energy_bins: array of fractions representing the spacing of
                the channels (for arbitrary channel structures)

        Returns:
            2-D array of energy values in keV for all samples
        """
        if fractional_energy_bins is None:
            fractional_energy_bins = np.linspace(0, 1, self.n_channels)

        all_channel_energies = []
        for i in range(self.n_samples):
            channel_energies = self.get_channel_energies(i, fractional_energy_bins)
            all_channel_energies.append(channel_energies)

        return all_channel_energies

    def get_channel_energies(self, sample_index, fractional_energy_bins=None) -> np.ndarray:
        """Get the energy (in keV) represented by the lower edge of each channel for a
        single sample at a specific index.

        See docstring for `_channels_to_energies()` for more details.

        Args:
            sample_index: index of the specific sample for which to obtain energies
            fractional_energy_bins: array of fractions representing the spacing of
                the channels (for arbitrary channel structures)

        Returns:
            Array of energy values in keV
        """
        if fractional_energy_bins is None:
            fractional_energy_bins = np.linspace(0, 1, self.n_channels)

        # TODO: raise error if ecal info is missing for any row

        offset, gain, quadratic, cubic, low_energy = self.ecal[sample_index]
        channel_energies = self._channels_to_energies(
            fractional_energy_bins,
            offset, gain, quadratic, cubic, low_energy
        )

        return channel_energies

    def get_confidences(self, fg_seeds_ss: SampleSet, bg_seed_ss: SampleSet = None,
                        bg_cps: float = None, is_lpe: bool = False,
                        confidence_func: Callable = None,
                        **confidence_func_kwargs) -> np.ndarray:
        """Get confidence measure for predictions as a NumPy array.

        Args:
            fg_seeds_ss: `Sampleset` containing foreground seeds
            bg_seed_ss: `Sampleset` containing a single background seed
            bg_cps: count rate used to scale `bg_seed_ss`
            is_lpe: whether predictions are for multi-class classification or label proportion
                estimation (LPE)
            confidence_func: metric function used to compare spectral reconstructions and
            confidence_func_kwargs: kwargs for confidence metric function

        Returns:
            Array containing confidence metric for each prediction
        """
        if fg_seeds_ss.spectra_type != SpectraType.Foreground:
            msg = (
                "`fg_seeds_ss` must have a `spectra_type` of `Foreground`. "
                f"Its `spectra_type` is `{fg_seeds_ss.spectra_type}`."
            )
            raise ValueError(msg)
        if self.spectra_type == SpectraType.Gross:
            if self.info.total_counts.isna().any() or (self.info.total_counts <= 0).any():
                msg = (
                    "This `SampleSet` must have the `info.total_counts` column filled in "
                    "with values greater than zero."
                    "If there are samples with spectra of all zeros, remove them first."
                )
                raise ValueError(msg)
            if self.info.live_time.isna().any() or (self.info.live_time <= 0).any():
                msg = (
                    "This `SampleSet` must have the `info.live_time` column filled in "
                    "with values greater than zero."
                    "If there are samples with live of zero, remove them first."
                )
                raise ValueError(msg)
            if not bg_seed_ss:
                raise ValueError("`bg_seed_ss` is required for current `spectra_type` of `Gross`.")
            if bg_seed_ss.n_samples != 1:
                raise ValueError("`bg_seed_ss` must have exactly one sample.")
            if bg_seed_ss.spectra_type != SpectraType.Background:
                msg = (
                    "`bg_seed_ss` must have a `spectra_type` of `Background`. "
                    f"Its `spectra_type` is `{bg_seed_ss.spectra_type}`."
                )
                raise ValueError(msg)
            if not bg_cps or bg_cps <= 0:
                raise ValueError("Positive background count rate required.")

        if confidence_func is None:
            from riid.losses import jensen_shannon_divergence
            confidence_func = jensen_shannon_divergence

        if is_lpe:
            recon_proportions = self.prediction_probas.values
        else:
            max_indices = np.argmax(self.prediction_probas.values, axis=1)
            recon_proportions = np.zeros(self.prediction_probas.values.shape)
            recon_proportions[np.arange(self.n_samples), max_indices] = 1

        reconstructions = np.dot(recon_proportions, fg_seeds_ss.spectra.values)

        if self.spectra_type == SpectraType.Gross:
            bg_spectrum = bg_seed_ss.spectra.iloc[0].values
            normalized_bg_spectrum = bg_spectrum / bg_spectrum.sum()

            bg_counts = bg_cps * self.info.live_time.values
            fg_counts = (self.info.total_counts.values - bg_counts).clip(1)

            scaled_bg_spectra = get_expected_spectra(normalized_bg_spectrum, bg_counts)
            scaled_fg_spectra = reconstructions * fg_counts[:, np.newaxis]

            reconstructions = scaled_fg_spectra + scaled_bg_spectra

        confidences = confidence_func(
            reconstructions.astype(np.float64),
            self.spectra.values,
            **confidence_func_kwargs
        )
        return np.array(confidences)

    def _get_spectral_distances(self, distance_func=distance.jensenshannon) -> np.array:
        n_samples = self.n_samples
        spectra = self.spectra.values.copy()
        spectrum_pairs = [
            (spectra[i], spectra[j])
            for i in range(n_samples)
            for j in range(i, n_samples)
        ]
        left_spectrum, right_spectrum = zip(*spectrum_pairs)
        distance_values = distance_func(left_spectrum, right_spectrum, axis=1)

        return distance_values

    def get_spectral_distance_matrix(self, distance_func=distance.jensenshannon,
                                     target_level="Seed") -> pd.DataFrame:
        """Compute the distance between all pairs of spectra.

        This method is intended for use on seed spectra (i.e., templates).
        Calling this method on large samplesets, such as those produced by the static
        synthesizer, is not recommended.
        This method only computes the upper triangle and diagonal of the matrix since the lower
        triangle would be a copy of the upper triangle.
        The diagonal is computed, for when Poisson noise is present, for two main reasons:

        - the distance between the same source will effectively never be zero.
        - there is sometimes research interest in comparing the diagonal to the upper triangle.

        Args:
            distance_func: specific distance function to use
            target_level: `SampleSet.sources` column level to use

        Returns:
            A 2-D array of distance values
        """
        distance_values = self._get_spectral_distances(distance_func)
        row_labels = self.get_labels(target_level=target_level)
        col_labels = self.sources.columns.get_level_values(level=target_level)
        distance_df = _get_distance_df_from_values(distance_values, row_labels, col_labels)

        return distance_df

    def get_labels(self, target_level="Isotope", max_only=True,
                   include_value=False, min_value=0.01,
                   level_aggregation=None):
        """Get row ground truth labels for each sample based on `sources` values.

        See docstring for `_get_row_labels()` for more details.
        """
        labels = _get_row_labels(
            self.sources,
            target_level=target_level,
            max_only=max_only,
            include_value=include_value,
            min_value=min_value,
            level_aggregation=level_aggregation,
        )
        return labels

    def get_predictions(self, target_level="Isotope", max_only=True,
                        include_value=False, min_value=0.01,
                        level_aggregation=None):
        """Get row predictions for each spectrum based on `prediction_probas` values.

        See docstring for `_get_row_labels()` for more details.
        """
        labels = _get_row_labels(
            self.prediction_probas,
            target_level=target_level,
            max_only=max_only,
            include_value=include_value,
            min_value=min_value,
            level_aggregation=level_aggregation,
        )
        return labels

    def get_samples(self) -> np.ndarray:
        """Get `spectra` as a NumPy array.

        Replaces NaN values with 0.

        Returns:
            Array containing spectrum data
        """
        spectra_values = np.nan_to_num(self._spectra)
        return spectra_values

    def get_source_contributions(self, target_level="Isotope") -> np.ndarray:
        """Get `sources` as a NumPy array.

        Replaces NaN values with 0.

        Returns:
            Array containing the ground truth contributions for each sample
        """
        collapsed_sources = self.sources.groupby(axis=1, level=target_level).sum()
        sources_values = np.nan_to_num(collapsed_sources)
        return sources_values

    def normalize(self, p: float = 1, clip_negatives: bool = True):
        """Apply L-p normalization to `spectra` in place.

        Default is L1 norm (p=1) can be interpreted as a forming a probability mass function (PMF).
        L2 norm (p=2) is unit energy (sum of squares == 1).

        Reference: https://en.wikipedia.org/wiki/Parseval's_theorem

        Args:
            p: exponent to which the spectra values are raised
            clip_negatives: whether negative values will be removed from the spectra and
                replaced with 0
        """
        if clip_negatives:
            self.clip_negatives()
        elif (self._spectra.values < 0).sum():
            logging.warning(
                "You are performing a normalization operation on spectra which contain negative " +
                "values.  Consider applying `clip_negatives`."
            )

        energy = (self._spectra.values ** p).sum(axis=1)[:, np.newaxis]
        energy[energy == 0] = 1
        self._spectra = pd.DataFrame(data=self._spectra.values / energy ** (1/p))
        if p == 1:
            self.spectra_state = SpectraState.L1Normalized
        elif p == 2:
            self.spectra_state = SpectraState.L2Normalized
        else:
            self.spectra_state = SpectraState.Unknown

    def normalize_sources(self):
        """Normalize `sources` such that rows sum to 1."""
        row_sums = self._sources.sum(axis=1)
        row_sums[row_sums == 0] = 1
        self._sources = self._sources.clip(0).divide(
            row_sums,
            axis=0,
        )

    def replace_nan(self, replace_value: float = 0):
        """Replace np.nan values.

        Args:
            replace_value: value with which NaN will be replaced

        """
        self._spectra.replace(np.nan, replace_value)
        self._sources.replace(np.nan, replace_value)
        self._info.replace(np.nan, replace_value)

    def sample(self, n_samples: int, random_seed: int = None) -> SampleSet:
        """Randomly sample the SampleSet.

        Args:
            n_samples: number of random samples to be returned
            random_seed: seed value for random number generation

        Returns:
            `Sampleset` of randomly selected measurements
        """
        if random_seed:
            random.seed(random_seed)

        if n_samples > self.n_samples:
            n_samples = self.n_samples

        random_indices = random.sample(self.spectra.index.values.tolist(), n_samples)
        random_mask = np.isin(self.spectra.index, random_indices)
        return self[random_mask]

    def set_dead_time_proportions(self):
        """Computes dead time proportion for each sample and saves it in the `SampleSet` info.
        """
        dead_time_props = self._get_dead_time_proportions()
        self.info[self.DEAD_TIME_PROP_INFO_KEY] = dead_time_props

    def shuffle(self, inplace: bool = True, random_state: int = None) -> SampleSet:
        """Randomly reorder all DataFrames.

        Args:
            inplace: whether to perform the operation in-place
            random_state: random seed for reproducing specific shuffles

        Returns:
            `SampleSet` object when `inplace` is False
        """
        new_row_ordering = np.arange(self.n_samples)
        np.random.default_rng(random_state).shuffle(new_row_ordering)

        new_ss = self if inplace else self[:]

        new_ss.spectra = self.spectra.iloc[new_row_ordering]\
            .reset_index(drop=True)
        new_ss.sources = self.sources.iloc[new_row_ordering]\
            .reset_index(drop=True)
        new_ss.info = self.info.iloc[new_row_ordering]\
            .reset_index(drop=True)
        if not new_ss.prediction_probas.empty:
            new_ss.prediction_probas = self.prediction_probas.iloc[new_row_ordering]\
                .reset_index(drop=True)

        if not inplace:
            return new_ss

    def sources_columns_to_dict(self, target_level="Isotope") -> Union[dict, list]:
        """Convert the MultiIndex columns of `sources` to a dictionary.

        Note: depending on `target_level` and `sources` columns, duplicate values are possible.

        Args:
            target_level: `SampleSet.sources` column level to use

        Returns:
            If `target_level` is "Category" or "Isotope," then a dict is returned.
            If `target_level` is "Seed," then a list is returned.

        Raises:
            `ValueError` when `target_level` is invalid
        """
        self._check_target_level(
            target_level,
            levels_allowed=SampleSet.SOURCES_MULTI_INDEX_NAMES
        )

        d = {}
        column_tuples = self.sources.columns.to_list()
        if target_level == "Seed":
            d = column_tuples
        elif target_level == "Isotope":
            for t in column_tuples:
                _, i, _ = t
                if i not in d:
                    d[i] = [t]
                if t not in d[i]:
                    d[i].append(t)
        else:  # target_level == "Category":
            for t in column_tuples:
                c, i, _ = t
                if c not in d:
                    d[c] = {}
                if i not in d[c]:
                    d[c][i] = [t]
                if t not in d[c][i]:
                    d[c][i].append(t)

        return d

    def split_fg_and_bg(self, bg_seed_names: Iterable = DEFAULT_BG_SEED_NAMES) \
            -> Tuple[SampleSet, SampleSet]:
        """Split the current `SampleSet` into two new `SampleSet`s, one containing only foreground
        sources and the other containing only background sources.

        Foreground sources are assumed to be anything that is not designated as a background source.

        Args:
            bg_seeds_names: names of the seeds which are considered background sources.
                This list be customized to also extract atypical background sources such as
                calibration sources.

        Returns:
            Two `SampleSet`s, the first containing only foregrounds and the second only
            containing backgrounds.
        """
        seed_labels = self.get_labels(target_level="Seed")
        bg_mask = seed_labels.isin(bg_seed_names)

        fg_seeds_ss = self[~bg_mask]
        fg_seeds_ss.drop_sources_columns_with_all_zeros()
        fg_seeds_ss.spectra_type = SpectraType.Foreground
        bg_seeds_ss = self[bg_mask]
        bg_seeds_ss.drop_sources_columns_with_all_zeros()
        bg_seeds_ss.spectra_type = SpectraType.Background

        return fg_seeds_ss, bg_seeds_ss

    def squash(self) -> SampleSet:
        """Combine all rows of the `SampleSet` into a single row.

        All data suited for summing is summed, otherwise the info from the first
        sample is used.

        Returns:
            A flattened `SampleSet`
        """
        flat_spectra = self.spectra.sum(axis=0).to_frame().T
        flat_sources = self.sources.sum(axis=0).to_frame().T
        flat_prediction_probas = self.prediction_probas.sum(axis=0).to_frame().T
        flat_info = self.info.iloc[0].to_frame().T

        if "description" in flat_info:
            flat_info.description = "squashed"
        if "live_time" in flat_info:
            flat_info.live_time = self.info.live_time.sum()
        if "real_time" in flat_info:
            flat_info.real_time = self.info.real_time.sum()
        if "total_counts" in flat_info:
            flat_info.total_counts = self.info.total_counts.sum()
        if "snr" in flat_info:
            flat_info.snr = self.info.snr.sum()

        # Create a new SampleSet with the flattened data
        flat_ss = SampleSet()
        flat_ss.spectra = flat_spectra
        flat_ss.sources = flat_sources
        flat_ss.prediction_probas = flat_prediction_probas
        flat_ss.info = flat_info
        # Copy dictionary and other non-DataFrame objects from full ss
        flat_ss._detector_info = self._detector_info
        flat_ss._synthesis_info = self._synthesis_info
        flat_ss._measured_or_synthetic = self._measured_or_synthetic

        return flat_ss

    def to_hdf(self, path: str, verbose=False, **kwargs):
        """Write the `SampleSet` to disk as a HDF file.

        Args:
            path: file path at which to save as an HDF file
            verbose: whether to display detailed output
            kwargs: additional arguments to be passed to the `Pandas.HDFStore` constructor

        Raises:
            `ValueError` when provided path extension is invalid
        """
        if not path.lower().endswith(riid.SAMPLESET_HDF_FILE_EXTENSION):
            logging.warning(f"Path does not end in {riid.SAMPLESET_HDF_FILE_EXTENSION}")

        _write_hdf(self, path, **kwargs)
        if verbose:
            logging.info(f"Saved SampleSet to '{path}'")

    def to_json(self, path: str, verbose=False):
        """Write the `SampleSet` to disk as a JSON file.

        Warning: it is not recommended that you use this on large `SampleSet` objects.
        Consider `SampleSet.to_hdf()` instead in such cases.

        Args:
            path: file path at which to save as an HDF file
            verbose: whether to display detailed output

        Raises:
            `ValueError` when provided path extension is invalid
        """
        if not path.lower().endswith(riid.SAMPLESET_JSON_FILE_EXTENSION):
            logging.warning(f"Path does not end in {riid.SAMPLESET_JSON_FILE_EXTENSION}")

        _write_json(self, path)
        if verbose:
            logging.info(f"Saved SampleSet to '{path}'")

    def to_pcf(self, path: str, verbose=True):
        """Write the `SampleSet` to disk as a PCF.

        Args:
            path: file path at which to save as a PCF
            verbose: whether to display detailed output

        Raises:
            `ValueError` when provided path extension is invalid
        """
        if not path.lower().endswith(riid.PCF_FILE_EXTENSION):
            logging.warning(f"Path does not end in {riid.PCF_FILE_EXTENSION}")

        _dict_to_pcf(_ss_to_pcf_dict(self, verbose), path, verbose)

        if verbose:
            logging.info(f"Saved SampleSet to '{path}'")

    def update_timestamp(self):
        """Set the timestamp for all samples to the current UTC date and time."""
        self.info.timestamp = _get_utc_timestamp()

    def upsample_spectra(self, target_bins: int = 4096):
        """Uniformly up-bin spectra.

        Args:
            target_bins: number of bins into which the current spectra
                should be split.
        """
        if target_bins % self.n_channels == 0:
            n_per_group = int(target_bins / self.n_channels)
            transformation = np.zeros([target_bins, self.n_channels])
            for i in range(self.n_channels):
                transformation[(i * n_per_group):((i * n_per_group) + n_per_group), i] = \
                                    1.0 / n_per_group
        else:
            raise ValueError(
                ("Desired number of bins ({}) is cannot"
                 " be uniformly mapped from existing"
                 " channels ({})").format(target_bins, self.n_channels))
        self._spectra = pd.DataFrame(
            data=np.matmul(
                self._spectra.values,
                transformation.T))


def read_hdf(path: str) -> SampleSet:
    """Read an HDF file in as a `SampleSet` object.

    Args:
        path: path for the HDF file to be read in

    Returns:
        `SampleSet` object
    """
    expanded_path = os.path.expanduser(path)
    if not os.path.isfile(expanded_path):
        raise FileNotFoundError(f"No file found at location '{expanded_path}'.")

    ss = _read_hdf(expanded_path)

    if not ss:
        raise FileNotFoundError(f"No data found at location '{expanded_path}'.")

    return ss


def read_pcf(path: str, verbose: bool = False) -> SampleSet:
    """Read a PCF file in as a `SampleSet` object.

    Args:
        path: path for the PCF file to be read in
        verbose: whether to show verbose function output in terminal

    Returns:
        `Sampleset` object
    """
    expanded_path = os.path.expanduser(path)
    return _pcf_dict_to_ss(_pcf_to_dict(expanded_path, verbose), verbose)


def _dict_to_bulleted_list(data_dict: dict, level=0, indent=4, bullet="-") -> str:
    lines = []
    level_key_width = max([len(x) for x in data_dict.keys()])
    for k, v in sorted(data_dict.items()):
        key = f"{k}:".ljust(level_key_width + 1)
        line = f"{' ' * (indent * level)}{bullet} {key}"
        if type(v) is dict:
            lines.append(line)
            sub_lines = _dict_to_bulleted_list(v, level + 1, indent, bullet)
            lines.extend(sub_lines)
        else:
            line += f" {v}"
            lines.append(line)
    return lines


def _get_utc_timestamp():
    ts = datetime.utcnow().isoformat(sep=" ", timespec="seconds")
    return ts


def _get_row_labels(df: pd.DataFrame, target_level: str = "Isotope", max_only: bool = True,
                    include_value: bool = False, min_value: float = 0.01,
                    level_aggregation: str = "sum") -> pd.Series:
    """Interprets the cell values in conjunction with the columns to determine an
    appropriate label for each row.

    Args:
        df: `DataFrame` with `MultiIndex` columns and named levels
        target_level: level of the columns to use for row labels
        max_only: for each row, whether to only use the column
            containing the max value.  When False, all column names
            containing a cell value >= `min_value` will be obtained.
        include_value: whether to include the cell value
            alongside the column name(s).
        min_value: when `max_only=False`, this filters out columns
            containing values less than this value. Note: this applies
            to values AFTER aggregation occurs.
        level_aggregation: method to use for combining
            cell values which have the same column name at the
            `target_level`.  Acceptable values are "sum" or "mean",
            otherwise this is ignored.

    Returns:
        Pandas `Series` where each entry is the row label
    """
    if max_only:
        if level_aggregation == "sum":
            values = df.groupby(axis=1, level=target_level).sum()
            labels = values.idxmax(axis=1)
        elif level_aggregation == "mean":
            values = df.groupby(axis=1, level=target_level).mean()
            labels = values.idxmax(axis=1)
        else:
            levels_to_drop = [
                x for x in SampleSet.SOURCES_MULTI_INDEX_NAMES
                if x != target_level and x in df.columns.names
            ]
            values = df.droplevel(levels_to_drop, axis=1)
            labels = values.idxmax(axis=1)
        if include_value:
            values = values.max(axis=1)
            labels = [f"{x} ({y:.2f})" for x, y in zip(labels, values)]
    else:  # Much slower
        if level_aggregation == "sum":
            values = df.groupby(axis=1, level=target_level).sum()
        elif level_aggregation == "mean":
            values = df.groupby(axis=1, level=target_level).mean()
        else:
            values = df
        mask = values.ge(min_value).values
        cols = values.columns.get_level_values(target_level).values
        if include_value:
            labels = [
                " + ".join([
                    f"{c} ({v:.2f})"
                    for c, v in zip(cols[x], values.iloc[i][x])
                ])
                for i, x in enumerate(mask)
            ]
        else:
            labels = [" + ".join(cols[x]) for x in mask]

    return pd.Series(labels)


def _validate_hdf_store_keys(keys: list):
    """Validates the `SampleSet` keys based on whether it contains the minimum required
    set of keys.

    Args:
        keys: collection of keys from the dataset being converted to a `SampleSet`

    Raises:
        `InvalidSampleSetFileError` when the set of keys provided does not include a required key
    """
    required_keys = [
        "/info",
        "/spectra",
        "/sources",
        "/prediction_probas"
    ]
    for rk in required_keys:
        if isinstance(rk, tuple):
            if all([ork not in keys for ork in rk]):
                msg = f"SampleSet did not contain at least one of the following pieces \
                        of required data: {rk}"
                raise InvalidSampleSetFileError(msg)
        else:
            if rk not in keys:
                msg = f"SampleSet did not contain the following piece of required data: {rk}"
                raise InvalidSampleSetFileError()


def _read_hdf(path: str) -> SampleSet:
    """Read `SampleSet` from an HDF file.

    Args:
        path: path where HDF file is to be read from

    Returns:
        `SampleSet` object
    """
    store = pd.HDFStore(path, mode="r")
    store_keys = store.keys()
    _validate_hdf_store_keys(store_keys)

    # Pull data from data store
    spectra = store.get("spectra")
    info = store.get("info")
    sources = store.get("sources")
    prediction_probas = store.get("prediction_probas")
    other_info = store.get("other_info")
    store.close()

    # Build SampleSet object
    ss = SampleSet()
    ss.spectra = spectra
    ss.sources = sources
    ss.info = info
    ss.prediction_probas = prediction_probas
    if "spectra_state" in other_info:
        ss.spectra_state = other_info["spectra_state"].iloc[0]
    if "measured_or_synthetic" in other_info:
        ss.measured_or_synthetic = other_info["measured_or_synthetic"].iloc[0]
    if "detector_info" in other_info:
        ss.detector_info = other_info["detector_info"].iloc[0]
    if "synthesis_info" in other_info:
        ss.synthesis_info = other_info["synthesis_info"].iloc[0]
    if "classified_by" in other_info:
        ss.classified_by = other_info["classified_by"].iloc[0]

    return ss


def _write_hdf(ss: SampleSet, output_path: str, **kwargs):
    """Write a `SampleSet` to an HDF file.

    Args:
        ss: `SampleSet` object to be written out
        output_path: path where file is to be written
        kwargs: additional arguments to be passed to the `Pandas.HDFStore` constructor
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        store_kwargs = {
            "complevel": 3,
            "complib": "blosc:blosclz",
        }
        store_kwargs.update(kwargs)

        put_kwargs = {
            "format": "fixed",
            "track_times": False,
        }
        with pd.HDFStore(output_path, "w", **store_kwargs) as store:
            store.put("spectra", ss.spectra, **put_kwargs)
            store.put("sources", ss.sources, **put_kwargs)
            store.put("info", ss.info, **put_kwargs)
            store.put("prediction_probas", ss.prediction_probas, **put_kwargs)

            other_info = {
                "spectra_state": ss.spectra_state,
                "measured_or_synthetic": ss.measured_or_synthetic,
                "detector_info": ss.detector_info,
                "synthesis_info": ss.synthesis_info,
                "classified_by": ss.classified_by,
            }
            store.put("other_info", pd.DataFrame(data=[other_info]), **put_kwargs)


def _ss_to_pcf_dict(ss: SampleSet, verbose=False) -> dict:
    """Convert a `SampleSet` to a dictionary representation.

    Args:
        ss: `SampleSet` object to convert to a dictionary
        verbose: whether to show detailed output

    Returns:
        Dictionary containing the values from the Sampleset
    """
    n_channels = int(ss.n_channels)
    n_records_per_spectrum = int((n_channels / 64) + 1)

    if "pcf_metadata" in ss.detector_info and ss.detector_info["pcf_metadata"]:
        pcf_header = ss.detector_info["pcf_metadata"]
    else:
        pcf_header = {
            "NRPS": n_records_per_spectrum,
            "Version": "DHS",
            "last_mod_hash": " " * 7,
            "UUID": " " * 36,
            "Inspection": " " * 16,
            "Lane_Number": 0,
            "Measurement_Remark": " " * 26,
            "Intrument_Type": " " * 28,
            "Manufacturer": " " * 28,
            "Instrument_Model": " " * 18,
            "Instrument_ID": " " * 18,
            "Item_Description": " " * 20,
            "Item_Location": " " * 16,
            "Measurement_Coordinates": " " * 16,
            "Item_to_detector_distance": 0,
            "Occupancy_Number": 0,
            "Cargo_Type": " " * 16,
            "SRSI": 83,
            "DevType": "DeviationPairsInFile",
        }

    spectra = []

    isotopes, seeds = pd.Series([], dtype=pd.StringDtype()), pd.Series([], dtype=pd.StringDtype())
    if not ss.sources.empty:
        isotope_level_name = SampleSet.SOURCES_MULTI_INDEX_NAMES[1]
        seed_level_names = SampleSet.SOURCES_MULTI_INDEX_NAMES[2]
        if isotope_level_name in ss.sources.columns.names:
            isotopes = ss.get_labels(target_level=isotope_level_name)
        if seed_level_names in ss.sources.columns.names:
            seeds = ss.get_labels(target_level=seed_level_names)

    sample_range = range(ss.n_samples)
    if verbose:
        sample_range = tqdm.tqdm(
            sample_range,
            desc="Converting sample set"
        )

    info_timestamps = ss.info.timestamp.fillna("")
    info_tags = ss.info.tag.fillna("")
    info_zero_fill = ss.info.fillna(0).astype(float, errors="ignore")
    for i in sample_range:
        title = isotopes[i] if not isotopes.empty else NO_ISOTOPE
        description = ss.info.description.fillna("").iloc[i]
        source = seeds[i] if not seeds.empty else NO_SEED
        compressed_text_buffer = _pack_compressed_text_buffer(title, description, source)

        header = {
            "Compressed_Text_Buffer": compressed_text_buffer,
            "Energy_Calibration_Offset": info_zero_fill.ecal_order_0.iloc[i],
            "Energy_Calibration_Gain": info_zero_fill.ecal_order_1.iloc[i],
            "Energy_Calibration_Quadratic": info_zero_fill.ecal_order_2.iloc[i],
            "Energy_Calibration_Cubic": info_zero_fill.ecal_order_3.iloc[i],
            "Energy_Calibration_Low_Energy": info_zero_fill.ecal_low_e.iloc[i],
            "Live_Time": info_zero_fill.live_time.iloc[i],
            "Total_time_per_real_time": info_zero_fill.real_time.iloc[i],
            "Number_of_Channels": n_channels,
            "Date-time_VAX": info_timestamps.iloc[i],
            "Occupancy_Flag": info_zero_fill.occupancy_flag.iloc[i],
            "Tag": info_tags.iloc[i],
            "Total_Neutron_Counts": info_zero_fill.neutron_counts.iloc[i],
        }

        spectrum = ss.spectra.iloc[i, :].values.astype(float).tolist()
        spectra.append({"header": header, "spectrum": spectrum})

    return {"header": pcf_header, "spectra": spectra}


def _pcf_dict_to_ss(pcf_dict: dict, verbose=True):
    """Convert a PCF dictionary into a `SampleSet`.

    Args:
        pcf_dict: dictionary representation of a PCF
        verbose: whether to display detailed output

    Returns:
        `Sampleset` object containing the PCF dict values
    """
    if not pcf_dict["spectra"]:
        return

    num_spectra = len(pcf_dict["spectra"])
    num_channels = len(pcf_dict["spectra"][0]["spectrum"])

    sources = []
    infos = []
    spectra = np.ndarray((num_spectra, num_channels))

    for i in range(0, num_spectra):
        spectrum = pcf_dict["spectra"][i]
        compressed_text_buffer = spectrum["header"]["Compressed_Text_Buffer"]

        title, description, source = _unpack_compressed_text_buffer(compressed_text_buffer)

        if not title and not source:
            source = NO_ISOTOPE
            title = NO_ISOTOPE
            category = NO_CATEGORY
        else:
            if not title:
                title = _find_isotope(source, verbose)
            elif not source:
                source = title
                title = _find_isotope(source, verbose)
            category = _find_category(title)

        # Extract any additional information from the source string
        distance_search = re.search("@ ([0-9,.]+)", source)
        if distance_search:
            distance = float(distance_search.group(1)) / 100
        else:
            distance = np.nan
        an_finds = re.findall("an=([0-9]+)", source)
        if an_finds:
            an = int(an_finds[0])
        else:
            an = None
        ad_finds = re.findall("ad=([0-9, .]+)", source)
        if ad_finds:
            ad_string = ad_finds[0].replace(",", "")
            ad = float(ad_string)
        else:
            ad = None

        order_0 = float(spectrum["header"]["Energy_Calibration_Offset"])
        order_1 = float(spectrum["header"]["Energy_Calibration_Gain"])
        order_2 = float(spectrum["header"]["Energy_Calibration_Quadratic"])
        order_3 = float(spectrum["header"]["Energy_Calibration_Cubic"])
        low_E = float(spectrum["header"]["Energy_Calibration_Low_Energy"])

        info = {
            "description": description,
            "timestamp": spectrum["header"]["Date-time_VAX"],
            "live_time": spectrum["header"]["Live_Time"],
            "real_time": spectrum["header"]["Total_time_per_real_time"],
            "total_counts": sum(spectrum["spectrum"]),
            "neutron_counts": spectrum["header"]["Total_Neutron_Counts"],
            "distance_cm": distance,
            "areal_density": ad,
            "ecal_order_0": order_0,
            "ecal_order_1": order_1,
            "ecal_order_2": order_2,
            "ecal_order_3": order_3,
            "ecal_low_e": low_E,
            "atomic_number": an,
            "occupancy_flag": spectrum["header"]["Occupancy_Flag"],
            "tag": spectrum["header"]["Tag"],
        }

        infos.append(info)
        sources.append((category, title, source))
        spectra[i, :] = np.array(spectrum["spectrum"])

    n_sources = len(sources)
    n_samples = len(spectra)
    new_index = pd.MultiIndex.from_tuples(sources, names=SampleSet.SOURCES_MULTI_INDEX_NAMES)
    sources_df = pd.DataFrame(np.zeros((n_samples, n_sources)), columns=new_index)
    for i, key in enumerate(sources):
        sources_df.loc[i, key] = 1.0

    ss = SampleSet()
    ss.spectra = pd.DataFrame(data=spectra)
    ss.sources = sources_df
    ss.info = pd.DataFrame(data=infos)
    ss.measured_or_synthetic = "synthetic",
    ss.detector_info["pcf_metadata"] = pcf_dict["header"]

    return ss


def _write_json(ss: SampleSet, output_path: str):
    ss_dict = _ss_to_pcf_dict(ss)
    ss_dict["detector_info"] = ss.detector_info
    with open(output_path, "w") as fout:
        json.dump(ss_dict, fout, indent=4)


def read_json(path: str) -> SampleSet:
    expanded_path = os.path.expanduser(path)
    with open(expanded_path, "r") as fin:
        ss_dict = json.load(fin)
    ss = _pcf_dict_to_ss(ss_dict)
    ss.detector_info = ss_dict["detector_info"]
    return ss


def _get_distance_df_from_values(distance_values: np.ndarray,
                                 row_labels: pd.Series, col_labels: pd.Series) -> pd.DataFrame:
    n_rows = row_labels.shape[0]
    n_cols = col_labels.shape[0]
    distance_df = pd.DataFrame(
        np.zeros((n_rows, n_cols)),
        index=row_labels,
        columns=col_labels,
    )
    n_values = len(distance_values)
    label_pairs = [
        (row_labels[i], col_labels[j])
        for i in range(n_rows)
        for j in range(i, n_cols)
    ]

    for i in range(n_values):
        value = distance_values[i]
        r, c = label_pairs[i]
        distance_df.at[r, c] = value

    return distance_df


class InvalidSampleSetFileError(Exception):
    """Missing or invalid keys in a file being read in as a `SampleSet`."""
    pass


class InvalidSampleCountError(Exception):
    """Incorrect number of samples encountered for the operation."""
    pass


class SpectraStateMismatchError(Exception):
    """Two `SampleSet`s have different `spectra_state` values."""
    pass


class ChannelCountMismatchError(Exception):
    """Two `SampleSet`s have different `n_channels` values."""
    pass
